"""Base dataset module."""

import abc
import tempfile
import urllib.parse
from pathlib import Path
from typing import Any, Optional, Union

import numpy as np  # type: ignore
import requests

from frouros.datasets.exceptions import (
    DownloadError,
    InvalidURLError,
    ReadFileError,
)
from frouros.utils.logger import logger


class BaseDatasetDownload(abc.ABC):
    """Abstract class representing a downloadable dataset."""

    def __init__(
        self,
        url: Union[str, list[str]],
        file_path: Optional[str] = None,
    ) -> None:
        """Init method.

        :param url: url or url mirrors from where dataset will be downloaded
        :type url: Union[str, list[str]]
        :param file_path: file path for the downloaded file
        :type file_path: str
        """
        self.url = url  # type: ignore
        self.file_path: Optional[Path] = (
            Path(file_path)
            if file_path
            else Path(tempfile.NamedTemporaryFile(delete=False).name)
        )

    @property
    def file_path(self) -> Optional[Path]:
        """File path property.

        :return: file path for the downloaded dataset
        :rtype: Optional[Path]
        """
        return self._file_path

    @file_path.setter
    def file_path(self, value: Optional[Path]) -> None:
        """File path setter.

        :param value: value to be set
        :type value: Optional[Path]
        """
        self._file_path = value

    @property
    def url(self) -> Union[str, list[str]]:
        """URL property.

        :return: URL from where dataset will be downloaded
        :rtype: Union[str, list[str]]
        """
        return self._url

    @url.setter
    def url(self, value: Union[str, list[str]]) -> None:
        """URL setter.

        :param value: value to be set
        :type value: Union[str, list[str]]
        :raises InvalidURLError: Invalid URL exception
        """
        urls = [value] if isinstance(value, str) else value
        for url in urls:
            if not self._check_valid_url(url=url):
                raise InvalidURLError(f"{value} is not a valid URL.")
        self._url = urls

    @staticmethod
    def _check_valid_url(url: str) -> bool:
        final_url = urllib.parse.urlparse(url=urllib.parse.urljoin(base=url, url="/"))
        is_valid = (
            all([final_url.scheme, final_url.netloc, final_url.path])
            and len(final_url.netloc.split(".")) > 1
        )
        return is_valid

    def _get_file(self, url: str) -> None:
        response = self._request_file(url=url)
        self._save_file(response=response)

    def _remove_temporal_file(self) -> None:
        self.file_path.unlink()  # type: ignore
        self.file_path = None

    def _request_file(self, url: str) -> requests.models.Response:
        logger.info("Trying to download data from %s to %s", url, self._file_path)
        request_head = requests.head(
            url=url,
            timeout=10,
        )
        if not request_head.ok:
            raise requests.exceptions.RequestException()
        request_response = requests.get(
            url=url,
            stream=True,
            timeout=10,
        )
        request_response.raise_for_status()
        return request_response

    def _save_file(self, response: requests.models.Response) -> None:
        try:
            self._write_file(content=response.content)
        except IOError as e:
            raise e

    def _write_file(self, content: bytes) -> None:
        with open(file=self.file_path, mode="ab") as f:  # type: ignore
            f.write(content)

    def download(self) -> None:
        """Download dataset.

        :raises DownloadError: Download exception
        """
        for url in self.url:
            try:
                self._get_file(url=url)
                break
            except requests.exceptions.RequestException:
                logger.warning("File cannot be downloaded from %s", url)
        else:
            raise DownloadError("File cannot be downloaded from any of the urls.")

    def load(self, **kwargs) -> Any:
        """Load dataset.

        :param kwargs: dict of kwargs
        :type kwargs: dict
        :raises FileNotFoundError: File not found exception
        :raises ReadFileError: Read file exception
        :return: loaded dataset
        :rtype: Any
        """
        if not self.file_path:
            raise FileNotFoundError(
                "Missing file to be loaded. "
                "Try using download() before "
                "load () method."
            )
        try:
            dataset = self.read_file(**kwargs)
        except IndexError as e:
            raise ReadFileError(e) from e
        self._remove_temporal_file()
        return dataset

    @abc.abstractmethod
    def read_file(self, **kwargs) -> Any:
        """Read file abstract method.

        :param kwargs: dict of kwargs
        :type kwargs: dict
        :return: read file
        :rtype: Any
        """

    def __repr__(self) -> str:
        """Repr method.

        :return: repr value
        :rtype: str
        """
        return (
            f"{self.__class__.__name__}"
            f"(url={self.url}, file_path='{self.file_path}')"
        )


class BaseDatasetGenerator(abc.ABC):  # noqa: B024
    """Abstract class representing a dataset generator."""

    def __init__(self, seed: Optional[int] = None) -> None:
        """Init method.

        :param seed: seed value
        :type seed: Optional[int]
        """
        try:
            np.random.seed(seed=seed)
        except (TypeError, ValueError) as e:
            raise e

    def __repr__(self) -> str:
        """Repr method.

        :return: repr value
        :rtype: str
        """
        return f"{self.__class__.__name__}()"



"""History callback module."""

from typing import Any, Optional, Union

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.utils.stats import BaseStat


class HistoryConceptDrift(BaseCallbackStreaming):
    """HistoryConceptDrift callback class that can be applied to :mod:`concept_drift.streaming <frouros.detectors.concept_drift.streaming>` detectors.

    :param name: name value, defaults to None. If None, the name will be set to `HistoryConceptDrift`.
    :type name: Optional[str]

    :Note:
    By default the following variables are stored:

    - `value`: list of values received by the detector
    - `drift`: list of drift flags
    - `num_instances`: list of number of instances received by the detector
    Each detector may store additional variables if they are defined in an `additional_vars` dictionary in the detectors `__init__` method.
    The user can add additional variables by calling the :func:`add_additional_vars` method.

    :Example:

    >>> from frouros.callbacks import HistoryConceptDrift
    >>> from frouros.detectors.concept_drift import ADWIN
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> dist_a = np.random.normal(loc=0.2, scale=0.01, size=1000)
    >>> dist_b = np.random.normal(loc=0.8, scale=0.04, size=1000)
    >>> stream = np.concatenate((dist_a, dist_b))
    >>> detector = ADWIN(callbacks=[HistoryConceptDrift(name="history")])
    >>> for i, value in enumerate(stream):
    ...     callbacks_log = detector.update(value=value)
    ...     if detector.drift:
    ...         print(f"Change detected at step {i}")
    ...         break
    Change detected at step 1055
    >>> callbacks_log["history"]["drift"]
    [False, False, ..., True]
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(name=name)
        self.additional_vars: list[str] = []
        self.history: dict[str, list[Any]] = {
            "value": [],
            "num_instances": [],
            "drift": [],
        }

    def add_additional_vars(self, vars_: list[str]) -> None:
        """Add additional variables to track.

        :param vars_: list of variables
        :type vars_: list[str]
        """
        self.additional_vars.extend(vars_)
        self.history = {**self.history, **{var: [] for var in self.additional_vars}}

    def on_update_end(self, value: Union[int, float]) -> None:
        """On update end method.

        :param value: value used to update the detector
        :type value: Union[int, float]
        """
        self.history["value"].append(value)
        self.history["num_instances"].append(
            self.detector.num_instances  # type: ignore
        )
        self.history["drift"].append(self.detector.drift)  # type: ignore
        for var in self.additional_vars:
            additional_var = self.detector.additional_vars[var]  # type: ignore
            # FIXME: Extract isinstance check to be done when  # pylint: disable=fixme
            #  add_addtional_vars is called (avoid the same computation)
            self.history[var].append(
                additional_var.get()
                if isinstance(additional_var, BaseStat)
                else additional_var
            )

        self.logs.update(**self.history)

    # FIXME: set_detector method as a workaround to  # pylint: disable=fixme
    #  avoid circular import problem. Make it an abstract method and
    #  uncomment commented code when it is solved

    # def set_detector(self, detector) -> None:
    #     """Set detector method.
    #
    #     :raises TypeError: Type error exception
    #     """
    #     if not isinstance(detector, BaseConceptDrift):
    #         raise TypeError(
    #             f"callback {self.__class__.name} cannot be used with detector"
    #             f" {detector.__class__name}. Must be used with a detector of "
    #             f"type BaseConceptDrift."
    #         )
    #     self.detector = detector

    def reset(self) -> None:
        """Reset method."""
        for key in self.history.keys():
            self.history[key].clear()



"""Base callback streaming module."""

import abc
from typing import Union

from frouros.callbacks.base import BaseCallback


class BaseCallbackStreaming(BaseCallback):
    """Callback streaming class."""

    def on_update_start(self, value: Union[int, float]) -> None:
        """On update start method.

        :param value: value used to update the detector
        :type value: Union[int, float]
        """

    def on_update_end(self, value: Union[int, float]) -> None:
        """On update end method.

        :param value: value used to update the detector
        :type value: Union[int, float]
        """

    # FIXME: set_detector method as a workaround to  # pylint: disable=fixme
    #  avoid circular import problem. Make it an abstract method and
    #  uncomment commented code when it is solved

    # @abc.abstractmethod
    # def set_detector(self, detector) -> None:
    #     """Set detector method."""

    @abc.abstractmethod
    def reset(self) -> None:
        """Reset method."""



"""Kernels module."""

import numpy as np  # type: ignore
from scipy.spatial.distance import cdist  # type: ignore


def rbf_kernel(
    X: np.ndarray, Y: np.ndarray, sigma: float = 1.0  # noqa: N803
) -> np.ndarray:
    """Radial basis function kernel between X and Y matrices.

    :param X: X matrix
    :type X: numpy.ndarray
    :param Y: Y matrix
    :type Y: numpy.ndarray
    :param sigma: sigma value (equivalent to gamma = 1 / (2 * sigma**2))
    :type sigma: float
    :return: Radial basis kernel matrix
    :rtype: numpy.ndarray
    """
    return np.exp(-cdist(X, Y, "sqeuclidean") / (2 * sigma**2))



"""Stats module."""

import abc
import itertools
from functools import partial
from multiprocessing import Pool
from typing import Any, Callable, Optional, Union, Tuple

import numpy as np  # type: ignore
from tqdm import tqdm  # type: ignore

from frouros.utils.data_structures import CircularQueue
from frouros.utils.logger import logger


class BaseStat(abc.ABC):
    """Abstract class representing an statistic."""

    @abc.abstractmethod
    def get(self) -> float:
        """Get method."""


class IncrementalStat(BaseStat):
    """Abstract class representing an incremental statistic."""

    @abc.abstractmethod
    def update(self, value: Union[int, float]) -> None:
        """Update abstract method."""

    @abc.abstractmethod
    def get(self) -> float:
        """Get method."""


class Mean(IncrementalStat):
    """Incremental mean class."""

    def __init__(  # noqa: D107
        self,
    ) -> None:
        self.mean = 0.0
        self.num_values = 0

    @property
    def mean(self) -> float:
        """Mean property.

        :return: mean value
        :rtype: float
        """
        return self._mean

    @mean.setter
    def mean(self, value: float) -> None:
        """Mean setter.

        :param value: value to be set
        :type value: float
        """
        self._mean = value

    @property
    def num_values(self) -> int:
        """Number of values property.

        :return: number of values
        :rtype: int
        """
        return self._num_values

    @num_values.setter
    def num_values(self, value: int) -> None:
        """Number of values setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("num_values must be greater of equal than 0.")
        self._num_values = value

    def update(self, value: Union[int, float]) -> None:
        """Update the mean value sequentially.

        :param value: value to use to update the mean
        :type value: int
        :raises TypeError: Type error exception
        """
        if not isinstance(value, (int, float, np.number)):
            raise TypeError("value must be of type int, float or numpy number.")
        self.num_values += 1
        self.mean += self.incremental_op(
            value=value,
            element=self.mean,
            size=self.num_values,
        )

    @staticmethod
    def incremental_op(
        value: Union[int, float],
        element: Union[int, float],
        size: int,
    ) -> float:
        """Incremental operation."""
        return (value - element) / size

    def get(self) -> float:
        """Get method."""
        return self.mean


class CircularMean(Mean):
    """Circular mean class.

    :param size: size of the circular mean
    :type size: int
    """

    def __init__(  # noqa: D107
        self,
        size: int,
    ) -> None:
        super().__init__()
        self.queue = CircularQueue(max_len=size)

    def update(self, value: Union[int, float]) -> None:
        """Update the mean value sequentially.

        :param value: value to use to update the mean
        :type value: int
        :raises TypeError: Type error exception
        """
        if not isinstance(value, (int, float, np.number)):
            raise TypeError("value must be of type int, float or numpy number.")
        element = self.queue.enqueue(value=value)
        self.num_values = len(self.queue)
        self.mean += self.incremental_op(
            value=value,
            element=self.mean if element is None else element,
            size=self.num_values,
        )


class EWMA(IncrementalStat):
    """EWMA (Exponential Weighted Moving Average) class.

    :param alpha: alpha value
    :type alpha: float
    """

    def __init__(  # noqa: D107
        self,
        alpha: float,
    ) -> None:
        self.alpha = alpha
        self.one_minus_alpha = 1.0 - self.alpha
        self.mean = 0

    @property
    def alpha(self) -> float:
        """Alpha property.

        :return: alpha value
        :rtype: float
        """
        return self._alpha

    @alpha.setter
    def alpha(self, value: float) -> None:
        """Alpha setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if not 0.0 <= value <= 1.0:
            raise ValueError("alpha must be in the range [0, 1].")
        self._alpha = value

    @property
    def mean(self) -> float:
        """Mean property.

        :return: mean value
        :rtype: float
        """
        return self._mean

    @mean.setter
    def mean(self, value: float) -> None:
        """Mean setter.

        :param value: value to be set
        :type value: float
        """
        self._mean = value

    def update(self, value: Union[int, float]) -> None:
        """Update the mean value sequentially.

        :param value: value to use to update the mean
        :type value: int
        :raises TypeError: Type error exception
        """
        if not isinstance(value, (int, float, np.number)):
            raise TypeError("value must be of type int, float or numpy number.")
        self.mean = self.alpha * value + self.one_minus_alpha * self.mean

    def get(self) -> float:
        """Get method."""
        return self.mean


def permutation(  # pylint: disable=too-many-arguments,too-many-locals
    X: np.ndarray,  # noqa: N803
    Y: np.ndarray,
    statistic: Callable,
    statistical_args: dict[str, Any],
    num_permutations: int,
    num_jobs: int,
    random_state: Optional[int] = None,
    verbose: bool = False,
) -> Tuple[list[float], int]:
    """Permutation method.

    :param X: reference data
    :type X: numpy.ndarray
    :param Y: test data
    :type Y: numpy.ndarray
    :param statistic: statistic to use
    :type statistic: Callable
    :param statistical_args: args to pass to statistic method
    :type statistical_args: dict[str, Any]
    :param num_permutations: number of permutations to use
    :type num_permutations: int
    :param num_jobs: number of jobs to use
    :type num_jobs: int
    :param random_state: random state value, defaults to None
    :type random_state: Optional[int]
    :param verbose: verbose flag, defaults to False
    :type verbose: bool
    :return: permuted statistics and max number of permutations
    :rtype: list[float], int
    """
    np.random.seed(seed=random_state)
    X_num_samples, Y_num_samples = X.shape[0], Y.shape[0]  # noqa: N806
    data = np.concatenate([X, Y])

    max_num_permutations = np.math.factorial(data.shape[0])
    if num_permutations >= max_num_permutations:
        logger.warning(
            "Number of permutations (%s) is greater or equal "
            "than the number of different possible permutations "
            "(%s). %s number of permutations will be used instead.",
            num_permutations,
            max_num_permutations,
            max_num_permutations,
        )
        permutations = np.array([*itertools.permutations(data)])
    else:
        permutations = [np.random.permutation(data) for _ in range(num_permutations)]
    permuted_data = []
    for data in permutations:
        permuted_data.append((data[:X_num_samples], data[-Y_num_samples:]))

    with Pool(processes=num_jobs) as pool:
        permuted_statistics = pool.starmap_async(
            partial(statistic, **statistical_args),
            iterable=tqdm(permuted_data) if verbose else permuted_data,
        ).get()

    return permuted_statistics, max_num_permutations



"""Test prequential error module."""

import numpy as np  # type: ignore
import pytest  # type: ignore

from frouros.metrics import PrequentialError


def error_scorer(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """Error scorer function.

    :param y_true: ground-truth values
    :type: numpy.ndarray
    :param y_pred: predicted values
    :type: numpy.ndarray
    :return error value
    :rtype float
    """
    return 1 - int(y_true == y_pred)


@pytest.mark.parametrize(
    "prequential_error, expected_performance",
    [
        (PrequentialError(alpha=1.0), 0.5),
        (
            PrequentialError(
                alpha=0.9999,
            ),
            0.50000833,
        ),
    ],
)
def test_prequential_error(
    prequential_error: PrequentialError, expected_performance: float
) -> None:
    """Test prequential error.

    :param prequential_error: prequential error metric
    :type prequential_error: PrequentialError
    :param expected_performance: expected performance value
    :type expected_performance: float
    """
    y_true = [True, True, False, True, False, True]
    y_pred = [True, False, False, False, True, True]
    for y_true_sample, y_pred_sample in zip(y_true, y_pred):
        error_value = error_scorer(
            y_true=np.array([y_true_sample]),
            y_pred=np.array([y_pred_sample]),
        )
        performance = prequential_error(error_value=error_value)

    assert performance == pytest.approx(expected_performance)



"""HDDM (Hoeffding's inequality drift detection method) module."""

import copy
from typing import Optional, Tuple, Union

import numpy as np  # type: ignore

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.concept_drift.streaming.statistical_process_control.base import (
    BaseSPCConfig,
    BaseSPC,
)
from frouros.utils.stats import EWMA, Mean


class BaseHDDMConfig(BaseSPCConfig):
    """HDDM (Hoeffding's drift detection method) [frias2014online]_ configuration.

    :References:

    .. [frias2014online] Frias-Blanco, Isvani, et al.
        "Online and non-parametric drift detection methods based on Hoeffding’s bounds."
        IEEE Transactions on Knowledge and Data Engineering 27.3 (2014): 810-823.
    """

    def __init__(
        self,
        alpha_d: float = 0.001,
        alpha_w: float = 0.005,
        two_sided_test: bool = False,
        min_num_instances: int = 30,
    ) -> None:
        """Init method.

        :param alpha_d: significance value for drift
        :type alpha_d: float
        :param alpha_w: significance value for warning
        :type alpha_w: float
        :param two_sided_test: flag that indicates if a two-sided test is performed
        :type two_sided_test: bool
        :param min_num_instances: minimum numbers of instances
        to start looking for changes
        :type min_num_instances: int
        """
        super().__init__(min_num_instances=min_num_instances)
        self.alpha_d = alpha_d
        self.alpha_w = alpha_w
        self.two_sided_test = two_sided_test

    @property
    def alpha_d(self) -> float:
        """Significance level d property.

        :return: significance level d
        :rtype: float
        """
        return self._alpha_d

    @alpha_d.setter
    def alpha_d(self, value: float) -> None:
        """Significance level d setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if not 0.0 < value <= 1.0:
            raise ValueError("alpha_d must be in the range (0, 1].")
        self._alpha_d = value

    @property
    def alpha_w(self) -> float:
        """Significance level w property.

        :return: significance level w
        :rtype: float
        """
        return self._alpha_w

    @alpha_w.setter
    def alpha_w(self, value: float) -> None:
        """Significance level w setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if not 0.0 < value <= 1.0:
            raise ValueError("alpha_w must be in the range (0, 1].")
        if value <= self.alpha_d:
            raise ValueError("alpha_w must be greater than alpha_d.")
        self._alpha_w = value

    @property
    def two_sided_test(self) -> bool:
        """Two-sided test flag property.

        :return: two-sided test flag
        :rtype: float
        """
        return self._two_sided_test

    @two_sided_test.setter
    def two_sided_test(self, value: bool) -> None:
        """Two-sided test flag setter.

        :param value: value to be set
        :type value: bool
        :raises ValueError: Value error exception
        """
        if not isinstance(value, bool):
            raise ValueError("two_sided_test must be of type bool.")
        self._two_sided_test = value


class HDDMAConfig(BaseHDDMConfig):
    """HDDM-A (Hoeffding's drift detection method A-Test) [frias2014online]_ configuration.

    :param alpha_d: significance value for drift, defaults to 0.001
    :type alpha_d: float
    :param alpha_w: significance value for warning, defaults to 0.005
    :type alpha_w: float
    :param two_sided_test: flag that indicates if a two-sided test is performed, defaults to False
    :type two_sided_test: bool
    :param min_num_instances: minimum numbers of instances to start looking for changes, defaults to 30
    :type min_num_instances: int

    :References:

    .. [frias2014online] Frias-Blanco, Isvani, et al.
        "Online and non-parametric drift detection methods based on Hoeffding’s bounds."
        IEEE Transactions on Knowledge and Data Engineering 27.3 (2014):
        810-823.
    """  # noqa: E501  pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        alpha_d: float = 0.001,
        alpha_w: float = 0.005,
        two_sided_test: bool = False,
        min_num_instances: int = 30,
    ) -> None:
        super().__init__(
            alpha_d=alpha_d,
            alpha_w=alpha_w,
            two_sided_test=two_sided_test,
            min_num_instances=min_num_instances,
        )


class HDDMWConfig(BaseHDDMConfig):
    """HDDM-W (Hoeffding's drift detection method W-Test) [frias2014online]_ configuration.

    :param alpha_d: significance value for drift, defaults to 0.001
    :type alpha_d: float
    :param alpha_w: significance value for warning, defaults to 0.005
    :type alpha_w: float
    :param two_sided_test: flag that indicates if a two-sided test is performed, defaults to False
    :type two_sided_test: bool
    :param lambda_: weight given to recent data compared to older data, defaults to 0.05
    :type lambda_: float
    :param min_num_instances: minimum numbers of instances to start looking for changes, defaults to 30
    :type min_num_instances: int

    :References:

    .. [frias2014online] Frias-Blanco, Isvani, et al.
        "Online and non-parametric drift detection methods based on Hoeffding’s bounds."
        IEEE Transactions on Knowledge and Data Engineering 27.3 (2014):
        810-823.
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        alpha_d: float = 0.001,
        alpha_w: float = 0.005,
        two_sided_test: bool = False,
        lambda_: float = 0.05,
        min_num_instances: int = 30,
    ) -> None:
        super().__init__(
            alpha_d=alpha_d,
            alpha_w=alpha_w,
            two_sided_test=two_sided_test,
            min_num_instances=min_num_instances,
        )
        self.lambda_ = lambda_

    @property
    def lambda_(self) -> float:
        """Weight recent data property.

        :return: weight given to recent data
        :rtype: float
        """
        return self._lambda_

    @lambda_.setter
    def lambda_(self, value: float) -> None:
        """Weight recent data setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if not 0.0 <= value <= 1.0:
            raise ValueError("lambda_ must be in the range [0, 1].")
        self._lambda_ = value


class HoeffdingOneSidedTest:
    """Hoeffding one-sided test class."""

    def __init__(self, alpha_d: float, alpha_w: float) -> None:
        """Init method."""
        self.alpha_d = alpha_d
        self.alpha_w = alpha_w
        self.x = Mean()
        self.z = Mean()

    def _check_mean_increase(self, m: int, alpha: float) -> bool:
        threshold = np.sqrt(
            m / (2 * self.x.num_values * self.z.num_values) * np.log(1 / alpha)
        )
        return self.z.mean - self.x.mean >= threshold

    def check_cases(self) -> Tuple[bool, bool]:
        """Check drift and warning cases.

        :return: drift and warning flags
        :rtype: Tuple[bool, bool]
        """
        m = self.z.num_values - self.x.num_values
        if self._check_mean_increase(m=m, alpha=self.alpha_d):
            drift = True
            warning = False
        elif self._check_mean_increase(m=m, alpha=self.alpha_w):
            drift = False
            warning = True
        else:
            drift = False
            warning = False

        return drift, warning

    def hoeffding_error_bound(self, num_values: int) -> float:
        """Calculate Hoeffding's error.

        :param num_values: number of values
        :type num_values: int
        :return: Hoeffding's error value
        :rtype: float
        """
        return np.sqrt(np.log(1 / self.alpha_d) / (2 * num_values))

    def reset(self) -> None:
        """Reset x and z incremental variables."""
        self.x = Mean()
        self.z = Mean()

    def set_initial_cut_mean(self) -> None:
        """Copy value of z to x if x has no values."""
        if self.x.num_values == 0:
            self.x = copy.deepcopy(self.z)

    def update_cut_point(self, epsilon_z: float) -> None:
        """Update cut point using Hoeffding's error.

        :param epsilon_z: Hoeffding's error of z
        :type epsilon_z: float
        """
        epsilon_x = self.hoeffding_error_bound(num_values=self.x.num_values)
        if self.z.mean + epsilon_z <= self.x.mean + epsilon_x:
            self.x = copy.deepcopy(self.z)


class HoeffdingTwoSidedTest(HoeffdingOneSidedTest):
    """Hoeffding two-sided test class."""

    def __init__(self, alpha_d: float, alpha_w: float) -> None:
        """Init method.

        :param alpha_d: significance value for drift
        :type alpha_d: float
        :param alpha_w: significance value for warning
        :type alpha_w: float
        """
        super().__init__(alpha_d=alpha_d, alpha_w=alpha_w)
        self.y = Mean()

    def reset(self) -> None:
        """Reset x, z and y incremental variables."""
        super().reset()
        self.y = Mean()

    def _check_mean_decrease(self, m: int, alpha: float) -> bool:
        threshold = np.sqrt(
            m / (2 * self.y.num_values * self.z.num_values) * np.log(1 / alpha)
        )
        return self.y.mean - self.z.mean >= threshold

    def check_cases(self) -> Tuple[bool, bool]:
        """Check drift and warning cases.

        :return: drift and warning flags
        :rtype: Tuple[bool, bool]
        """
        drift_increase, warning_increase = super().check_cases()
        m = self.z.num_values - self.y.num_values
        if self._check_mean_decrease(m=m, alpha=self.alpha_d):
            drift_decrease = True
            warning_decrease = False
        elif self._check_mean_decrease(m=m, alpha=self.alpha_w):
            drift_decrease = False
            warning_decrease = True
        else:
            drift_decrease = False
            warning_decrease = False
        return any([drift_increase, drift_decrease]), any(
            [warning_increase, warning_decrease]
        )

    def set_initial_cut_mean(self) -> None:
        """Copy value of z to x and/or y if x and/or y has no values."""
        super().set_initial_cut_mean()
        if self.y.num_values == 0:
            self.y = copy.deepcopy(self.z)

    def update_cut_point(self, epsilon_z: float) -> None:
        """Update cut point using Hoeffding's error.

        :param epsilon_z: Hoeffding's error of z
        :type epsilon_z: float
        """
        super().update_cut_point(epsilon_z=epsilon_z)
        epsilon_y = self.hoeffding_error_bound(num_values=self.y.num_values)
        if self.y.mean - epsilon_y <= self.z.mean - epsilon_z:
            self.y = copy.deepcopy(self.z)


class HDDMA(BaseSPC):
    """HDDM-A (Hoeffding's drift detection method with A-Test) [frias2014online]_ detector.

    :param config: configuration object of the detector, defaults to None. If None, the default configuration of :class:`HDDMAConfig` is used.
    :type config: Optional[HDDMAConfig]
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]]

    :Note:
    :func:`update` method expects to receive a value of 0 if the instance is correctly classified (no error) and 1 otherwise (error).

    :References:

    .. [frias2014online] Frias-Blanco, Isvani, et al.
        "Online and non-parametric drift detection methods based on Hoeffding’s bounds."
        IEEE Transactions on Knowledge and Data Engineering 27.3 (2014):
        810-823.

    :Example:

    >>> from frouros.detectors.concept_drift import HDDMA
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> dist_a = np.random.binomial(n=1, p=0.6, size=1000)
    >>> dist_b = np.random.binomial(n=1, p=0.8, size=1000)
    >>> stream = np.concatenate((dist_a, dist_b))
    >>> detector = HDDMA()
    >>> warning_flag = False
    >>> for i, value in enumerate(stream):
    ...     _ = detector.update(value=value)
    ...     if detector.drift:
    ...         print(f"Change detected at step {i}")
    ...         break
    ...     if not warning_flag and detector.warning:
    ...         print(f"Warning detected at step {i}")
    ...         warning_flag = True
    Warning detected at step 1043
    Change detected at step 1054
    """  # noqa: E501  # pylint: disable=line-too-long

    config_type = HDDMAConfig  # type: ignore

    def __init__(  # noqa: D107
        self,
        config: Optional[HDDMAConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        super().__init__(
            config=config,
            callbacks=callbacks,
        )
        self.additional_vars = {
            "test_type": (
                HoeffdingTwoSidedTest(
                    alpha_d=self.config.alpha_d,  # type: ignore
                    alpha_w=self.config.alpha_w,  # type: ignore
                )
                if self.config.two_sided_test  # type: ignore
                else HoeffdingOneSidedTest(
                    alpha_d=self.config.alpha_d,  # type: ignore
                    alpha_w=self.config.alpha_w,  # type: ignore
                )
            ),
            **self.additional_vars,  # type: ignore
        }
        self._set_additional_vars_callback()

    @property
    def test_type(self) -> HoeffdingOneSidedTest:
        """Test type property.

        :return: test type value
        :rtype: HoeffdingOneSidedTest
        """
        return self._additional_vars["test_type"]

    def _update(self, value: Union[int, float], **kwargs) -> None:
        self.num_instances += 1

        self.test_type.z.update(value=value)
        self.test_type.set_initial_cut_mean()

        epsilon_z = self.test_type.hoeffding_error_bound(
            num_values=self.test_type.z.num_values
        )
        self.test_type.update_cut_point(epsilon_z=epsilon_z)

        if self.num_instances >= self.config.min_num_instances:
            drift_flag, warning_flag = (
                self.test_type.check_cases()
                if self.test_type.z.num_values != self.test_type.x.num_values
                else (False, False)
            )
            if drift_flag:
                # Out-of-Control
                self.drift = True
                self.warning = False
                self.test_type.reset()
            else:
                if warning_flag:
                    # Warning
                    self.warning = True
                else:
                    # In-Control
                    self.warning = False
                self.drift = False
        else:
            self.drift, self.warning = False, False


class SampleInfo:
    """Sample information class."""

    def __init__(self, lambda_: float) -> None:
        """Init method.

        :param lambda_: weight given to recent data compared to older data
        :type lambda_: float
        """
        self.ewma = EWMA(alpha=lambda_)
        self.independent_bound_condition = 1.0
        self.lambda_squared = lambda_ * lambda_
        self.one_minus_lambda_squared = (
            self.ewma.one_minus_alpha * self.ewma.one_minus_alpha
        )

    @property
    def ewma(self) -> EWMA:
        """EWMA (Exponential Moving Average) property.

        :return: EWMA value
        :rtype: EWMA
        """
        return self._ewma

    @ewma.setter
    def ewma(self, value: EWMA) -> None:
        """EWMA (Exponential Moving Average)  setter.

        :param value: value to be set
        :type value: EWMA
        :raises TypeError: Type error exception
        """
        if not isinstance(value, EWMA):
            raise TypeError("ewma must be of type EWMA.")
        self._ewma = value

    def update(self, value: float) -> None:
        """Update statistics.

        :param value: value to use in update
        :type value: float
        """
        self.ewma.update(value=value)
        self.independent_bound_condition = (
            self.lambda_squared
            + self.one_minus_lambda_squared * self.independent_bound_condition
        )


class McDiarmidOneSidedTest:
    """McDiarmid one-sided test class."""

    def __init__(self, alpha_d: float, alpha_w: float, lambda_: float) -> None:
        """Init method.

        :param alpha_d: significance value for drift
        :type alpha_d: float
        :param alpha_w: significance value for warning
        :type alpha_w: float
        :param lambda_: weight given to recent data compared to older data
        :type lambda_: float
        """
        self.alpha_d = alpha_d
        self.alpha_w = alpha_w
        self.lambda_ = lambda_
        self.total = SampleInfo(lambda_=self.lambda_)
        self.sample_increase_1 = SampleInfo(lambda_=self.lambda_)
        self.sample_increase_2 = SampleInfo(lambda_=self.lambda_)
        self.increase_cut_point = float("inf")

    def _check_threshold(
        self, sample_1: SampleInfo, sample_2: SampleInfo, alpha: float
    ) -> bool:
        independent_bound_condition_sum = (
            sample_1.independent_bound_condition + sample_2.independent_bound_condition
        )
        threshold = self._mcdiarmid_error_bound(
            independent_bound_condition=independent_bound_condition_sum,
            alpha=alpha,
        )
        return sample_2.ewma.mean - sample_1.ewma.mean > threshold

    def _check_mean_increase(self, alpha: float) -> bool:
        return self._check_threshold(
            sample_1=self.sample_increase_1,
            sample_2=self.sample_increase_2,
            alpha=alpha,
        )

    def check_changes(self) -> Tuple[bool, bool]:
        """Check drift and warning cases.

        :return: drift and warning flags
        :rtype: Tuple[bool, bool]
        """
        drift = self._check_mean_increase(alpha=self.alpha_d)
        warning = False if drift else self._check_mean_increase(alpha=self.alpha_w)
        return drift, warning

    @staticmethod
    def _mcdiarmid_error_bound(
        independent_bound_condition: float, alpha: float
    ) -> float:
        return np.sqrt(independent_bound_condition * np.log(1 / alpha) / 2)

    def reset(self) -> None:
        """Reset sample info variables and increase cut point."""
        self.total = SampleInfo(lambda_=self.lambda_)
        self.sample_increase_1 = SampleInfo(lambda_=self.lambda_)
        self.sample_increase_2 = SampleInfo(lambda_=self.lambda_)
        self.increase_cut_point = float("inf")

    def update_stats(self, value: float, alpha: float) -> None:
        """Update statistics.

        :param value: value to use in update
        :type value: float
        :param alpha: significance value
        :type alpha: float
        """
        self.total.update(value=value)
        epsilon = self._mcdiarmid_error_bound(
            independent_bound_condition=self.total.independent_bound_condition,
            alpha=alpha,
        )
        ewma_plus_epsilon = self.total.ewma.mean + epsilon
        if ewma_plus_epsilon < self.increase_cut_point:
            self.increase_cut_point = ewma_plus_epsilon
            self.sample_increase_1 = copy.deepcopy(self.total)
            self.sample_increase_2 = SampleInfo(lambda_=self.lambda_)
        else:
            self.sample_increase_2.update(value=value)


class McDiarmidTwoSidedTest(McDiarmidOneSidedTest):
    """McDiarmid two-sided test class."""

    def __init__(self, alpha_d: float, alpha_w: float, lambda_: float) -> None:
        """Init method.

        :param alpha_d: significance value for drift
        :type alpha_d: float
        :param alpha_w: significance value for warning
        :type alpha_w: float
        :param lambda_: weight given to recent data compared to older data
        :type lambda_: float
        """
        super().__init__(alpha_d=alpha_d, alpha_w=alpha_w, lambda_=lambda_)
        self.sample_decrease_1 = SampleInfo(lambda_=lambda_)
        self.sample_decrease_2 = SampleInfo(lambda_=lambda_)
        self.decrease_cut_point = float("-inf")

    def _check_mean_decrease(self, alpha: float) -> bool:
        return self._check_threshold(
            sample_1=self.sample_increase_2,
            sample_2=self.sample_increase_1,
            alpha=alpha,
        )

    def check_changes(self) -> Tuple[bool, bool]:
        """Check drift and warning cases.

        :return: drift and warning flags
        :rtype: Tuple[bool, bool]
        """
        drift_increase, warning_increase = super().check_changes()
        drift_decrease = (
            False if drift_increase else self._check_mean_decrease(alpha=self.alpha_d),
        )
        warning_decrease = (
            False
            if warning_increase or drift_decrease
            else self._check_mean_decrease(alpha=self.alpha_w)
        )
        return any([drift_increase, drift_decrease]), any(
            [warning_increase, warning_decrease]
        )

    def reset(self) -> None:
        """Reset sample info variables and cut points."""
        super().reset()
        self.sample_decrease_1 = SampleInfo(lambda_=self.lambda_)
        self.sample_decrease_2 = SampleInfo(lambda_=self.lambda_)
        self.decrease_cut_point = float("-inf")

    def update_stats(self, value: float, alpha: float) -> None:
        """Update statistics.

        :param value: value to use in update
        :type value: float
        :param alpha: significance value
        :type alpha: float
        """
        super().update_stats(value=value, alpha=alpha)
        epsilon = self._mcdiarmid_error_bound(
            independent_bound_condition=self.total.independent_bound_condition,
            alpha=alpha,
        )
        ewma_minus_epsilon = self.total.ewma.mean - epsilon
        if ewma_minus_epsilon > self.decrease_cut_point:
            self.decrease_cut_point = ewma_minus_epsilon
            self.sample_decrease_1 = copy.deepcopy(self.total)
            self.sample_decrease_2 = SampleInfo(lambda_=self.lambda_)
        else:
            self.sample_decrease_2.update(value=value)


class HDDMW(BaseSPC):
    """HDDM-W (Hoeffding's drift detection method with W-Test) [frias2014online]_ detector.

    :param config: configuration object of the detector, defaults to None. If None, the default configuration of :class:`HDDMWConfig` is used.
    :type config: Optional[HDDMWConfig]
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]]

    :Note:
    :func:`update` method expects to receive a value of 0 if the instance is correctly classified (no error) and 1 otherwise (error).

    :References:

    .. [frias2014online] Frias-Blanco, Isvani, et al.
        "Online and non-parametric drift detection methods based on Hoeffding’s bounds."
        IEEE Transactions on Knowledge and Data Engineering 27.3 (2014):
        810-823.

    :Example:

    >>> from frouros.detectors.concept_drift import HDDMW
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> dist_a = np.random.binomial(n=1, p=0.6, size=1000)
    >>> dist_b = np.random.binomial(n=1, p=0.8, size=1000)
    >>> stream = np.concatenate((dist_a, dist_b))
    >>> detector = HDDMW()
    >>> warning_flag = False
    >>> for i, value in enumerate(stream):
    ...     _ = detector.update(value=value)
    ...     if detector.drift:
    ...         print(f"Change detected at step {i}")
    ...         break
    ...     if not warning_flag and detector.warning:
    ...         print(f"Warning detected at step {i}")
    ...         warning_flag = True
    Warning detected at step 1017
    Change detected at step 1029
    """  # noqa: E501  # pylint: disable=line-too-long

    config_type = HDDMWConfig  # type: ignore

    def __init__(  # noqa: D107
        self,
        config: Optional[HDDMWConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        super().__init__(
            config=config,
            callbacks=callbacks,
        )
        self.additional_vars = {
            "test_type": (
                McDiarmidTwoSidedTest(
                    alpha_d=self.config.alpha_d,  # type: ignore
                    alpha_w=self.config.alpha_w,  # type: ignore
                    lambda_=self.config.lambda_,  # type: ignore
                )
                if self.config.two_sided_test  # type: ignore
                else McDiarmidOneSidedTest(
                    alpha_d=self.config.alpha_d,  # type: ignore
                    alpha_w=self.config.alpha_w,  # type: ignore
                    lambda_=self.config.lambda_,  # type: ignore
                )
            ),
            **self.additional_vars,  # type: ignore
        }
        self._set_additional_vars_callback()

    @property
    def test_type(self) -> McDiarmidOneSidedTest:
        """Test type property.

        :return: test type value
        :rtype: McDiarmidOneSidedTest
        """
        return self._additional_vars["test_type"]

    def _update(self, value: Union[int, float], **kwargs) -> None:
        self.num_instances += 1

        self.test_type.update_stats(
            value=value, alpha=self.config.lambda_  # type: ignore
        )

        if self.num_instances >= self.config.min_num_instances:
            drift_flag, warning_flag = self.test_type.check_changes()
            if drift_flag:
                # Out-of-Control
                self.drift = True
                self.warning = False
                self.test_type.reset()
            else:
                if warning_flag:
                    # Warning
                    self.warning = True
                else:
                    # In-Control
                    self.warning = False
                self.drift = False
        else:
            self.drift, self.warning = False, False



"""Base concept drift module."""

import abc
from typing import Any, Optional, Union

from frouros.callbacks import HistoryConceptDrift
from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.base import BaseDetector
from frouros.utils.checks import check_callbacks


class BaseConceptDriftConfig(abc.ABC):  # noqa: B024
    """Abstract class representing a concept drift configuration class."""

    def __init__(
        self,
        min_num_instances: int = 10,
    ) -> None:
        """Init method.

        :param min_num_instances: minimum numbers of instances
        to start looking for changes
        :type min_num_instances: int
        """
        self.min_num_instances = min_num_instances

    @property
    def min_num_instances(self) -> int:
        """Minimum number of instances property.

        :return: minimum number of instances to start looking for changes
        :rtype: int
        """
        return self._min_num_instances

    @min_num_instances.setter
    def min_num_instances(self, value: int) -> None:
        """Minimum number of instances setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 1:
            raise ValueError("value must be greater than 0.")
        self._min_num_instances = value

    def __repr__(self) -> str:
        """Repr method.

        :return: repr value
        :rtype: str
        """
        return (
            f"{self.__class__.__name__}"
            f"({', '.join(f'{k[1:]}={v}' for k, v in self.__dict__.items())})"
        )


class BaseConceptDrift(BaseDetector):
    """Abstract class representing a concept drift streaming."""

    config_type = BaseConceptDriftConfig

    def __init__(
        self,
        config: Optional[BaseConceptDriftConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        """Init method.

        :param config: configuration parameters
        :type config: Optional[BaseConceptDriftConfig]
        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallbackStreaming,
        list[BaseCallbackStreaming]]]
        """
        check_callbacks(
            callbacks=callbacks,
            expected_cls=BaseCallbackStreaming,  # type: ignore
        )
        super().__init__(callbacks=callbacks)  # type: ignore
        self.config = config  # type: ignore
        self.additional_vars = None

        self.num_instances = 0
        self.drift = False
        for callback in self.callbacks:  # type: ignore
            callback.set_detector(detector=self)

    def _set_additional_vars_callback(self) -> None:
        for callback in self.callbacks:  # type: ignore
            if isinstance(callback, HistoryConceptDrift):
                # callback.set_detector(detector=self)
                callback.add_additional_vars(
                    vars_=self.additional_vars.keys(),  # type: ignore
                )

    @property
    def additional_vars(self) -> Optional[dict[str, Any]]:
        """Additional variables property.

        :return: additional variables
        :rtype: Optional[dict[str, Any]]
        """
        return self._additional_vars

    @additional_vars.setter
    def additional_vars(self, value: Optional[dict[str, Any]]) -> None:
        """Additional variables setter.

        :param value: value to be set
        :type value: Optional[dict[str, Any]]
        """
        self._additional_vars = value if value is not None else {}

    @property
    def config(self) -> BaseConceptDriftConfig:
        """Config property.

        :return: configuration parameters of the estimator
        :rtype: BaseConceptDriftConfig
        """
        return self._config

    @config.setter
    def config(self, value: Optional[BaseConceptDriftConfig]) -> None:
        """Config setter.

        :param value: value to be set
        :type value: BaseConceptDriftConfig
        :raises TypeError: Type error exception
        """
        if value is not None:
            if not isinstance(value, self.config_type):
                raise TypeError(
                    f"value must be of type " f"{self.config_type.__class__.__name__}."
                )
            self._config = value
        else:
            self._config = self.config_type()

    @property
    def num_instances(self) -> int:
        """Number of instances counter property.

        :return: Number of instances counter value
        :rtype: int
        """
        return self._num_instances

    @num_instances.setter
    def num_instances(self, value: int) -> None:
        """Number of instances counter setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("num_instances must be greater or equal than 0.")
        self._num_instances = value

    def reset(self) -> None:
        """Reset method."""
        self.num_instances = 0
        self.drift = False
        for callback in self.callbacks:  # type: ignore
            callback.reset()

    @property
    def status(self) -> dict[str, bool]:
        """Status property.

        :return: status dict
        :rtype: dict[str, bool]
        """
        return {"drift": self.drift}

    def update(self, value: Union[int, float], **kwargs) -> dict[str, Any]:
        """Update method.

        :param value: value to update detector
        :type value: Union[int, float]
        :return: callbacks logs
        :rtype: dict[str, Any]]
        """
        for callback in self.callbacks:  # type: ignore
            callback.on_update_start(  # type: ignore
                value=value,
            )
        self._update(value=value, **kwargs)
        for callback in self.callbacks:  # type: ignore
            callback.on_update_end(  # type: ignore
                value=value,
            )

        callbacks_logs = self._get_callbacks_logs()
        return callbacks_logs

    def _get_callbacks_logs(self) -> dict[str, Any]:
        logs = {
            callback.name: callback.logs for callback in self.callbacks  # type: ignore
        }
        return logs

    @abc.abstractmethod
    def _update(self, value: Union[int, float], **kwargs) -> None:
        pass

    def __repr__(self) -> str:
        """Repr method.

        :return: repr value
        :rtype: str
        """
        return f"{super().__repr__()[:-1]}, config={self.config})"



"""Data drift batch detection methods init."""

from .distance_based import (
    BhattacharyyaDistance,
    EMD,
    EnergyDistance,
    HellingerDistance,
    HINormalizedComplement,
    JS,
    KL,
    PSI,
    MMD,
)
from .statistical_test import (
    AndersonDarlingTest,
    BWSTest,
    ChiSquareTest,
    CVMTest,
    KSTest,
    KuiperTest,
    MannWhitneyUTest,
    WelchTTest,
)

__all__ = [
    "AndersonDarlingTest",
    "BWSTest",
    "BhattacharyyaDistance",
    "ChiSquareTest",
    "CVMTest",
    "EMD",
    "EnergyDistance",
    "HellingerDistance",
    "HINormalizedComplement",
    "JS",
    "KL",
    "KSTest",
    "KuiperTest",
    "PSI",
    "MannWhitneyUTest",
    "MMD",
    "WelchTTest",
]



"""Base concept drift streaming module."""

import abc
from typing import Union

from frouros.detectors.concept_drift.base import (
    BaseConceptDrift,
    BaseConceptDriftConfig,
)


class BaseConceptDriftStreamingConfig(BaseConceptDriftConfig):
    """Abstract class representing a concept drift streaming configuration class."""


class BaseConceptDriftStreaming(BaseConceptDrift):
    """Abstract class representing a concept drift streaming detector."""

    @abc.abstractmethod
    def _update(self, value: Union[int, float], **kwargs) -> None:
        pass



"""Base data drift batch module."""

import abc
from typing import Any, Optional, Tuple, Union

import numpy as np  # type: ignore

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.data_drift.base import (
    BaseDataDrift,
    BaseDataType,
    BaseResult,
    BaseStatisticalType,
)
from frouros.utils.checks import check_callbacks


class BaseDataDriftStreaming(BaseDataDrift):
    """Abstract class representing a data drift streaming detector."""

    def __init__(
        self,
        data_type: BaseDataType,
        statistical_type: BaseStatisticalType,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        """Init method.

        :param data_type: data type
        :type data_type: BaseDataType
        :param statistical_type: statistical type
        :type statistical_type: BaseStatisticalType
        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallbackStreaming],
        list[BaseCallbackStreaming]]
        """
        check_callbacks(
            callbacks=callbacks,
            expected_cls=BaseCallbackStreaming,  # type: ignore
        )
        super().__init__(
            callbacks=callbacks,  # type: ignore
            data_type=data_type,
            statistical_type=statistical_type,
        )
        self.num_instances = 0
        for callback in self.callbacks:  # type: ignore
            callback.set_detector(detector=self)

    def reset(self) -> None:
        """Reset method."""
        super().reset()
        self.num_instances = 0
        self._reset()

    def update(
        self,
        value: Union[int, float],
    ) -> Tuple[Optional[BaseResult], dict[str, Any]]:
        """Update detector.

        :param value: value to use to update the detector
        :type value: Union[int, float]
        :return: update result and callbacks logs
        :rtype: Tuple[Optional[BaseResult], dict[str, Any]]
        """
        self._common_checks()  # noqa: N806
        self._specific_checks(X=value)  # noqa: N806
        self.num_instances += 1

        for callback in self.callbacks:  # type: ignore
            callback.on_update_start(  # type: ignore
                value=value,  # type: ignore
            )
        result = self._update(value=value)
        for callback in self.callbacks:  # type: ignore
            callback.on_update_end(  # type: ignore
                value=result,  # type: ignore
            )

        callbacks_logs = self._get_callbacks_logs()
        return result, callbacks_logs

    def _specific_checks(self, X: np.ndarray) -> None:  # noqa: N803
        pass

    @abc.abstractmethod
    def _fit(self, X: np.ndarray) -> None:  # noqa: N803
        pass

    @abc.abstractmethod
    def _reset(self) -> None:
        pass

    @abc.abstractmethod
    def _update(self, value: Union[int, float]) -> Optional[BaseResult]:
        pass



"""MMD (Maximum Mean Discrepancy) module."""

import itertools
from typing import Callable, Generator, Optional, Union

import numpy as np  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.base import MultivariateData
from frouros.detectors.data_drift.batch.distance_based.base import (
    BaseDistanceBased,
    DistanceResult,
)
from frouros.utils.kernels import rbf_kernel


class MMD(BaseDistanceBased):
    """MMD (Maximum Mean Discrepancy) [gretton2012kernel]_ detector.

    :param kernel: kernel function, defaults to :func:`rbf_kernel() <frouros.utils.kernels.rbf_kernel>`
    :type kernel: Callable
    :param chunk_size: chunk size value, defaults to None
    :type chunk_size: Optional[int]
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]]

    :References:

    .. [gretton2012kernel] Gretton, Arthur, et al.
        "A kernel two-sample test."
        The Journal of Machine Learning Research 13.1 (2012): 723-773.

    :Example:

    >>> from functools import partial
    >>> from frouros.detectors.data_drift import MMD
    >>> from frouros.utils.kernels import rbf_kernel
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> X = np.random.multivariate_normal(mean=[1, 1], cov=[[2, 0], [0, 2]], size=100)
    >>> Y = np.random.multivariate_normal(mean=[0, 0], cov=[[2, 1], [1, 2]], size=100)
    >>> detector = MMD(kernel=partial(rbf_kernel, sigma=0.5))
    >>> _ = detector.fit(X=X)
    >>> detector.compare(X=Y)[0]
    DistanceResult(distance=0.02146955300299802)
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        kernel: Callable = rbf_kernel,
        chunk_size: Optional[int] = None,
        callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]] = None,
    ) -> None:
        super().__init__(
            statistical_type=MultivariateData(),
            statistical_method=self._mmd,
            statistical_kwargs={
                "kernel": kernel,
                "chunk_size": chunk_size,
            },
            callbacks=callbacks,
        )
        self.kernel = kernel
        self.chunk_size = chunk_size
        self._expected_k_xx = None

    @property
    def chunk_size(self) -> Optional[int]:
        """Chunk size property.

        :return: chunk size to use
        :rtype: int
        """
        return self._chunk_size

    @chunk_size.setter
    def chunk_size(self, value: Optional[int]) -> None:
        """Chunk size method setter.

        :param value: value to be set
        :type value: Optional[int]
        :raises TypeError: Type error exception
        """
        if value is not None:
            if isinstance(value, int):  # type: ignore
                if value <= 0:
                    raise ValueError("chunk_size must be greater than 0 or None.")
            else:
                raise TypeError("chunk_size must be of type int or None.")
        self._chunk_size = value

    @property
    def kernel(self) -> Callable:
        """Kernel property.

        :return: kernel function to use
        :rtype: Callable
        """
        return self._kernel

    @kernel.setter
    def kernel(self, value: Callable) -> None:
        """Kernel method setter.

        :param value: value to be set
        :type value: Callable
        :raises TypeError: Type error exception
        """
        if not isinstance(value, Callable):  # type: ignore
            raise TypeError("kernel must be of type Callable.")
        self._kernel = value

    def _distance_measure(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,  # noqa: N803
        **kwargs,
    ) -> DistanceResult:
        mmd = self._mmd(
            X=X_ref,
            Y=X,
            kernel=self.kernel,
            chunk_size=self.chunk_size,
            expected_k_xx=self._expected_k_xx,
            **kwargs,
        )
        distance_test = DistanceResult(distance=mmd)
        return distance_test

    def _fit(
        self,
        X: np.ndarray,  # noqa: N803
    ) -> None:
        super()._fit(X=X)
        # Add dimension only for the kernel calculation (if dim == 1)
        if X.ndim == 1:
            X = np.expand_dims(X, axis=1)  # noqa: N806
        x_num_samples = len(self.X_ref)  # type: ignore

        chunk_size_x = (
            x_num_samples
            if self.chunk_size is None
            else self.chunk_size  # type: ignore
        )

        x_chunks = self._get_chunks(  # noqa: N806
            data=X,
            chunk_size=chunk_size_x,
        )
        x_chunks_combinations = itertools.product(x_chunks, repeat=2)  # noqa: N806

        k_xx_sum = (
            self._compute_kernel(
                chunk_combinations=x_chunks_combinations,  # type: ignore
                kernel=self.kernel,
            )
            # Remove diagonal (j!=i case)
            - x_num_samples
        )

        self._expected_k_xx = k_xx_sum / (  # type: ignore
            x_num_samples * (x_num_samples - 1)
        )

    @staticmethod
    def _compute_kernel(chunk_combinations: Generator, kernel: Callable) -> float:
        k_sum = np.array([kernel(*chunk).sum() for chunk in chunk_combinations]).sum()
        return k_sum

    @staticmethod
    def _get_chunks(data: np.ndarray, chunk_size: int) -> Generator:
        chunks = (
            data[i : i + chunk_size]  # noqa: E203
            for i in range(0, len(data), chunk_size)
        )
        return chunks

    @staticmethod
    def _mmd(  # pylint: disable=too-many-locals
        X: np.ndarray,  # noqa: N803
        Y: np.ndarray,
        *,
        kernel: Callable,
        **kwargs,
    ) -> float:  # noqa: N803
        # Only check for X dimension (X == Y dim comparison has been already made)
        if X.ndim == 1:
            X = np.expand_dims(X, axis=1)  # noqa: N806
            Y = np.expand_dims(Y, axis=1)  # noqa: N806

        x_num_samples = len(X)  # noqa: N806
        chunk_size_x = (
            kwargs["chunk_size"]
            if "chunk_size" in kwargs and kwargs["chunk_size"] is not None
            else x_num_samples
        )

        # If expected_k_xx is provided, we don't need to compute it again
        if "expected_k_xx" in kwargs:
            x_chunks_copy = MMD._get_chunks(  # noqa: N806
                data=X,
                chunk_size=chunk_size_x,
            )
            expected_k_xx = kwargs["expected_k_xx"]
        else:
            # Compute expected_k_xx
            x_chunks, x_chunks_copy = itertools.tee(  # type: ignore
                MMD._get_chunks(
                    data=X,
                    chunk_size=chunk_size_x,
                ),
                2,
            )
            x_chunks_combinations = itertools.product(  # type: ignore
                x_chunks,
                repeat=2,
            )
            k_xx_sum = (
                MMD._compute_kernel(
                    chunk_combinations=x_chunks_combinations,  # type: ignore
                    kernel=kernel,
                )
                # Remove diagonal (j!=i case)
                - x_num_samples
            )
            expected_k_xx = k_xx_sum / (x_num_samples * (x_num_samples - 1))

        y_num_samples = len(Y)  # noqa: N806
        chunk_size_y = (
            kwargs["chunk_size"]
            if "chunk_size" in kwargs and kwargs["chunk_size"] is not None
            else y_num_samples
        )
        y_chunks, y_chunks_copy = itertools.tee(  # noqa: N806
            MMD._get_chunks(
                data=Y,
                chunk_size=chunk_size_y,
            ),
            2,
        )
        y_chunks_combinations = itertools.product(  # noqa: N806
            y_chunks,
            repeat=2,
        )
        xy_chunks_combinations = itertools.product(  # noqa: N806
            x_chunks_copy,
            y_chunks_copy,
        )

        k_yy_sum = (
            MMD._compute_kernel(
                chunk_combinations=y_chunks_combinations,  # type: ignore
                kernel=kernel,
            )
            # Remove diagonal (j!=i case)
            - y_num_samples
        )
        k_xy_sum = MMD._compute_kernel(
            chunk_combinations=xy_chunks_combinations,  # type: ignore
            kernel=kernel,
        )
        mmd = (
            +expected_k_xx
            + k_yy_sum / (y_num_samples * (y_num_samples - 1))
            - 2 * k_xy_sum / (x_num_samples * y_num_samples)
        )
        return mmd



"""STEPD (Statistical test of equal proportions) module."""

from typing import Optional, Union

import numpy as np  # type: ignore
from scipy.stats import norm  # type: ignore

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.concept_drift.streaming.window_based.base import (
    BaseWindowConfig,
    BaseWindow,
)
from frouros.utils.data_structures import AccuracyQueue


class STEPDConfig(BaseWindowConfig):
    """STEPD (Statistical test of equal proportions) [nishida2007detecting]_ configuration.

    :param alpha_d: significance value for overall, defaults to 0.003
    :type alpha_d: float
    :param alpha_w: significance value for last, defaults to 0.05
    :type alpha_w: float
    :param min_num_instances: minimum numbers of instances to start looking for changes, defaults to 30
    :type min_num_instances: int

    :References:

    .. [nishida2007detecting] Nishida, Kyosuke, and Koichiro Yamauchi.
        "Detecting concept drift using statistical testing." Discovery science.
        Vol. 4755. 2007.
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        alpha_d: float = 0.003,
        alpha_w: float = 0.05,
        min_num_instances: int = 30,
    ) -> None:
        super().__init__(min_num_instances=min_num_instances)
        self.alpha_d = alpha_d
        self.alpha_w = alpha_w

    @property
    def alpha_d(self) -> float:
        """Significance level d property.

        :return: significance level d
        :rtype: float
        """
        return self._alpha_d

    @alpha_d.setter
    def alpha_d(self, value: float) -> None:
        """Significance level d setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value <= 0.0:
            raise ValueError("alpha_d must be greater than 0.0.")
        self._alpha_d = value

    @property
    def alpha_w(self) -> float:
        """Significance level w property.

        :return: significance level w
        :rtype: float
        """
        return self._alpha_w

    @alpha_w.setter
    def alpha_w(self, value: float) -> None:
        """Significance level w setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value <= 0.0:
            raise ValueError("alpha_w must be greater than 0.0.")
        if value <= self.alpha_d:
            raise ValueError("alpha_w must be greater than alpha_d.")
        self._alpha_w = value


class STEPD(BaseWindow):
    """STEPD (Statistical test of equal proportions) [nishida2007detecting]_ detector.

    :param config: configuration object of the detector, defaults to None. If None, the default configuration of :class:`STEPDConfig` is used.
    :type config: Optional[STEPDConfig]
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]]

    :References:

    .. [nishida2007detecting] Nishida, Kyosuke, and Koichiro Yamauchi.
        "Detecting concept drift using statistical testing." Discovery science.
        Vol. 4755. 2007.

    :Example:

    >>> from frouros.detectors.concept_drift import STEPD, STEPDConfig
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> dist_a = np.random.binomial(n=1, p=0.8, size=1000)
    >>> dist_b = np.random.binomial(n=1, p=0.5, size=1000)
    >>> stream = np.concatenate((dist_a, dist_b))
    >>> detector = STEPD(config=STEPDConfig(alpha_d=0.001, alpha_w=0.005))
    >>> for i, value in enumerate(stream):
    ...     _ = detector.update(value=value)
    ...     if detector.drift:
    ...         print(f"Change detected at step {i}")
    ...         break
    ...     if detector.warning:
    ...         print(f"Warning detected at step {i}")
    Warning detected at step 640
    Warning detected at step 641
    Warning detected at step 1023
    Change detected at step 1024
    """  # noqa: E501  # pylint: disable=line-too-long

    config_type = STEPDConfig  # type: ignore

    def __init__(  # noqa: D107
        self,
        config: Optional[STEPDConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        super().__init__(
            config=config,
            callbacks=callbacks,
        )
        self.additional_vars = {
            "correct_total": 0,
            # FIXME include get method in AccuracyQueue  # pylint: disable=fixme
            "window_accuracy": AccuracyQueue(max_len=self.config.min_num_instances),
            **self.additional_vars,  # type: ignore
        }
        self.warning = False
        self._set_additional_vars_callback()
        self._min_num_instances = 2 * self.config.min_num_instances
        self._distribution = norm()

    @property
    def warning(self) -> bool:
        """Warning property.

        :return: warning
        :rtype: bool
        """
        return self._warning

    @warning.setter
    def warning(self, value: bool) -> None:
        """Warning setter.

        :param value: value to be set
        :type value: bool
        """
        self._warning = value

    @property
    def correct_total(self) -> int:
        """Number of correct labels property.

        :return: accuracy scorer function
        :rtype: int
        """
        return self._additional_vars["correct_total"]

    @correct_total.setter
    def correct_total(self, value: int) -> None:
        """Number of correct labels setter.

        :param value: value to be set
        :type value: int
        """
        self._additional_vars["correct_total"] = value

    @property
    def num_instances_overall(self) -> int:
        """Number of overall instances property.

        :return: number of overall instances
        :rtype: int
        """
        return self.num_instances - self.num_instances_window

    @property
    def correct_overall(self) -> int:
        """Number of correct overall labels property.

        :return: number of overall labels
        :rtype: int
        """
        return self.correct_total - self.correct_window

    @property
    def correct_window(self) -> int:
        """Number of correct window labels property.

        :return: number of window labels
        :rtype: int
        """
        return self.window_accuracy.num_true

    @property
    def num_instances_window(self) -> int:
        """Number of window instances property.

        :return: number of window instances
        :rtype: int
        """
        return self.window_accuracy.size

    @property
    def window_accuracy(self) -> AccuracyQueue:
        """Accuracy window property.

        :return: accuracy window
        :rtype: AccuracyQueue
        """
        return self._additional_vars["window_accuracy"]

    @window_accuracy.setter
    def window_accuracy(self, value: AccuracyQueue) -> None:
        """Accuracy window setter.

        :param value: value to be set
        :type value: AccuracyQueue
        :raises TypeError: Type error exception
        """
        if not isinstance(value, AccuracyQueue):
            raise TypeError("value must be of type AccuracyQueue")
        self._additional_vars["window_accuracy"] = value

    def _calculate_statistic(self):
        p_hat = self.correct_total / self.num_instances
        num_instances_inv = (
            1 / self.num_instances_overall + 1 / self.num_instances_window
        )
        statistic = (
            np.abs(
                self.correct_overall / self.num_instances_overall
                - self.correct_window / self.num_instances_window
            )
            - 0.5 * num_instances_inv
        ) / np.sqrt((p_hat * (1 - p_hat) * num_instances_inv))
        return statistic

    def reset(self) -> None:
        """Reset method."""
        super().reset()
        self.correct_total = 0
        self.window_accuracy.clear()

    def _update(self, value: Union[int, float], **kwargs) -> None:
        self.num_instances += 1

        self.correct_total += np.sum(value)
        self.window_accuracy.enqueue(value=value)

        if self.num_instances >= self._min_num_instances:
            statistic = self._calculate_statistic()
            p_value = self._distribution.sf(np.abs(statistic))  # One-sided test

            if p_value < self.config.alpha_d:  # type: ignore
                # Drift case
                self.drift = True
                self.warning = False
            else:
                if p_value < self.config.alpha_w:  # type: ignore
                    # Warning case
                    self.warning = True
                else:
                    # In-Control
                    self.warning = False
                self.drift = False
        else:
            self.drift, self.warning = False, False



"""Base callback batch module."""

import abc
from typing import Any

import numpy as np  # type: ignore

from frouros.callbacks.base import BaseCallback


class BaseCallbackBatch(BaseCallback):
    """Callback batch class."""

    def on_compare_start(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X_test: np.ndarray,
    ) -> None:
        """On compare start method.

        :param X_ref: reference data
        :type X_ref: numpy.ndarray
        :param X_test: test data
        :type X_test: numpy.ndarray
        """

    def on_compare_end(
        self,
        result: Any,
        X_ref: np.ndarray,  # noqa: N803
        X_test: np.ndarray,
    ) -> None:
        """On compare end method.

        :param result: result obtained from the `compare` method
        :type result: Any
        :param X_ref: reference data
        :type X_ref: numpy.ndarray
        :param X_test: test data
        :type X_test: numpy.ndarray
        """

    # FIXME: set_detector method as a workaround to  # pylint: disable=fixme
    #  avoid circular import problem. Make it an abstract method and
    #  uncomment commented code when it is solved

    # @abc.abstractmethod
    # def set_detector(self, detector) -> None:
    #     """Set detector method."""

    @abc.abstractmethod
    def reset(self) -> None:
        """Reset method."""



"""Data drift batch statistical test detection methods' init."""

from .anderson_darling import AndersonDarlingTest
from .bws import BWSTest
from .chisquare import ChiSquareTest
from .cvm import CVMTest
from .ks import KSTest
from .kuiper_test import KuiperTest
from .mann_whitney_u import MannWhitneyUTest
from .welch_t_test import WelchTTest

__all__ = [
    "AndersonDarlingTest",
    "BWSTest",
    "ChiSquareTest",
    "CVMTest",
    "KSTest",
    "KuiperTest",
    "MannWhitneyUTest",
    "WelchTTest",
]



"""Concept drift exception module."""


class InvalidAverageRunLengthError(Exception):
    """Invalid average run length exception."""


class NoFitMethodError(Exception):
    """Not fit method exception."""


class UpdateDetectorError(Exception):
    """Update detector exception."""



"""Data drift statistical test base module."""

import abc
from typing import Optional, Union

import numpy as np  # type: ignore

from frouros.detectors.data_drift.base import BaseResult
from frouros.detectors.data_drift.streaming.base import (
    BaseDataDriftStreaming,
)


class StatisticalResult(BaseResult):
    """Statistical result class."""

    def __init__(
        self,
        statistic: Union[int, float],
        p_value: Union[int, float],
    ) -> None:
        """Init method.

        :param statistic: statistic value
        :type statistic: Union[int, float]
        :param p_value: p-value
        :type p_value: Union[int, float]
        """
        super().__init__()
        self.statistic = statistic
        self.p_value = p_value

    @property
    def statistic(self) -> Union[int, float]:
        """Statistic value property.

        :return: statistic value
        :rtype: Union[int, float]
        """
        return self._statistic

    @statistic.setter
    def statistic(self, value: Union[int, float]) -> None:
        """Statistic value setter.

        :param value: value to be set
        :type value: Union[int, float]
        """
        self._statistic = value

    @property
    def p_value(self) -> Union[int, float]:
        """P-value property.

        :return: p-value
        :rtype: Union[int, float]
        """
        return self._p_value

    @p_value.setter
    def p_value(self, value: Union[int, float]) -> None:
        """P-value setter.

        :param value: value to be set
        :type value: Union[int, float]
        """
        if not 0 <= value <= 1:
            raise ValueError("p-value must be between 0 and 1.")
        self._p_value = value


class BaseStatisticalTest(BaseDataDriftStreaming):
    """Abstract class representing a statistical test."""

    @abc.abstractmethod
    def _fit(self, X: np.ndarray) -> None:  # noqa: N803
        pass

    @abc.abstractmethod
    def _reset(self) -> None:
        pass

    @abc.abstractmethod
    def _update(self, value: Union[int, float]) -> Optional[StatisticalResult]:
        pass

    @staticmethod
    @abc.abstractmethod
    def _statistical_test(
        X_ref: np.ndarray, X: np.ndarray, **kwargs  # noqa: N803
    ) -> StatisticalResult:
        pass



"""Concept drift window based detection methods' init."""

from .adwin import ADWIN, ADWINConfig
from .kswin import KSWIN, KSWINConfig
from .stepd import STEPD, STEPDConfig


__all__ = [
    "ADWIN",
    "ADWINConfig",
    "KSWIN",
    "KSWINConfig",
    "STEPD",
    "STEPDConfig",
]



"""Geometric Moving Average module."""

from typing import Optional, Union

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.concept_drift.streaming.change_detection.base import (
    BaseCUSUM,
    BaseCUSUMConfig,
    AlphaConfig,
)


class GeometricMovingAverageConfig(BaseCUSUMConfig, AlphaConfig):
    """Geometric Moving Average [robertst1959control]_ configuration.

    :param alpha: forgetting factor value, defaults to 0.99
    :type alpha: float
    :param lambda_: delta value, defaults to 1.0
    :type lambda_: float
    :param min_num_instances: minimum numbers of instances to start looking for changes, defaults to 30
    :type min_num_instances: int

    :References:

    .. [robertst1959control] Roberts, S. W.
        “Control Chart Tests Based on Geometric Moving Averages.”
        Technometrics, vol. 1, no. 3, 1959, pp. 239–50.
        JSTOR, https://doi.org/10.2307/1266443.
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        alpha: float = 0.99,
        lambda_: float = 1.0,
        min_num_instances: int = 30,
    ) -> None:
        BaseCUSUMConfig.__init__(
            self, lambda_=lambda_, min_num_instances=min_num_instances
        )
        AlphaConfig.__init__(self, alpha=alpha)


class GeometricMovingAverage(BaseCUSUM):
    """Geometric Moving Average [robertst1959control]_ detector.

    :param config: configuration object of the detector, defaults to None. If None, the default configuration of :class:`GeometricMovingAverageConfig` is used.
    :type config: Optional[GeometricMovingAverageConfig]
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]]

    :References:

    .. [robertst1959control] Roberts, S. W.
        “Control Chart Tests Based on Geometric Moving Averages.”
        Technometrics, vol. 1, no. 3, 1959, pp. 239–50.
        JSTOR, https://doi.org/10.2307/1266443.

    :Example:

    >>> from frouros.detectors.concept_drift import GeometricMovingAverage, GeometricMovingAverageConfig
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> dist_a = np.random.normal(loc=0.2, scale=0.01, size=1000)
    >>> dist_b = np.random.normal(loc=0.8, scale=0.04, size=1000)
    >>> stream = np.concatenate((dist_a, dist_b))
    >>> detector = GeometricMovingAverage(config=GeometricMovingAverageConfig(lambda_=0.3))
    >>> for i, value in enumerate(stream):
    ...     _ = detector.update(value=value)
    ...     if detector.drift:
    ...         print(f"Change detected at step {i}")
    ...         break
    Change detected at step 1071
    """  # noqa: E501  # pylint: disable=line-too-long

    config_type = GeometricMovingAverageConfig  # type: ignore

    def __init__(  # noqa: D107
        self,
        config: Optional[GeometricMovingAverageConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        super().__init__(
            config=config,
            callbacks=callbacks,
        )

    def _update_sum(self, error_rate: float) -> None:
        self.sum_ = self.config.alpha * self.sum_ + (  # type: ignore
            1 - self.config.alpha  # type: ignore
        ) * (error_rate - self.mean_error_rate.mean)



"""MMD (Maximum Mean Discrepancy) module."""

from typing import Any, Callable, Optional, Tuple, Union

import numpy as np  # type: ignore

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.data_drift.base import NumericalData, MultivariateData
from frouros.detectors.data_drift.batch import MMD as MMDBatch  # noqa: N811
from frouros.detectors.data_drift.batch.distance_based.mmd import rbf_kernel
from frouros.detectors.data_drift.streaming.distance_based.base import (
    BaseDistanceBased,
    DistanceResult,
)
from frouros.utils.data_structures import CircularQueue


class MMD(BaseDistanceBased):
    """MMD (Maximum Mean Discrepancy) [gretton2012kernel]_ detector.

    :param window_size: window size value
    :type window_size: int
    :param kernel: kernel function, defaults to :func:`rbf_kernel() <frouros.utils.kernels.rbf_kernel>`
    :type kernel: Callable
    :param chunk_size: chunk size value, defaults to None
    :type chunk_size: Optional[int]
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackStreaming,
    list[BaseCallbackStreaming]]]

    :References:

    .. [gretton2012kernel] Gretton, Arthur, et al.
        "A kernel two-sample test."
        The Journal of Machine Learning Research 13.1 (2012): 723-773.

    :Example:

    >>> from functools import partial
    >>> from frouros.detectors.data_drift import MMDStreaming
    >>> from frouros.utils.kernels import rbf_kernel
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> X = np.random.multivariate_normal(mean=[1, 1], cov=[[2, 0], [0, 2]], size=100)
    >>> Y = np.random.multivariate_normal(mean=[0, 0], cov=[[2, 1], [1, 2]], size=100)
    >>> detector = MMDStreaming(window_size=10, kernel=partial(rbf_kernel, sigma=0.5))
    >>> _ = detector.fit(X=X)
    >>> for sample in Y:
    ...     distance, _ = detector.update(value=sample)
    ...     if distance is not None:
    ...         print(distance)
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        window_size: int,
        kernel: Callable = rbf_kernel,
        chunk_size: Optional[int] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        super().__init__(
            data_type=NumericalData(),
            statistical_type=MultivariateData(),
            callbacks=callbacks,
        )
        self.mmd = MMDBatch(
            kernel=kernel,
            chunk_size=chunk_size,
        )
        self.window_size = window_size
        self.X_queue = CircularQueue(max_len=self.window_size)

    @property
    def window_size(self) -> int:
        """Window size property.

        :return: window size
        :rtype: int
        """
        return self._window_size

    @window_size.setter
    def window_size(self, value: int) -> None:
        """Window size setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 1:
            raise ValueError("window_size value must be greater than 0.")
        self._window_size = value

    def _fit(self, X: np.ndarray) -> None:  # noqa: N803
        self.mmd.fit(X=X)
        self.X_ref = self.mmd.X_ref

    def _reset(self) -> None:
        super()._reset()
        self.mmd.reset()

    def _update(self, value: Union[int, float]) -> Optional[DistanceResult]:
        self.X_queue.enqueue(value=value)

        if self.num_instances < self.window_size:
            return None

        # FIXME: Handle callback logs. Now are ignored.  # pylint: disable=fixme
        distance, _ = self.mmd.compare(X=np.array(self.X_queue))
        return distance

    def _compare(
        self,
        X: np.ndarray,  # noqa: N803
    ) -> Tuple[Optional[DistanceResult], dict[str, Any]]:  # noqa: N803
        return self.mmd.compare(X=X)



"""Base concept drift ChangeDetection based module."""

import abc
from typing import Optional, Union

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.concept_drift.streaming.base import (
    BaseConceptDriftStreaming,
    BaseConceptDriftStreamingConfig,
)
from frouros.utils.stats import Mean


class BaseChangeDetectionConfig(BaseConceptDriftStreamingConfig):
    """Class representing a ChangeDetection based configuration class."""


class BaseChangeDetection(BaseConceptDriftStreaming):
    """ChangeDetection based algorithm class."""

    config_type = BaseChangeDetectionConfig

    @abc.abstractmethod
    def _update(self, value: Union[int, float], **kwargs) -> None:
        pass


class BaseCUSUMConfig(BaseChangeDetectionConfig):
    """Class representing a CUSUM based configuration class.

    :param lambda_: lambda value, defaults to 50.0
    :type lambda_: float
    :param min_num_instances: minimum numbers of instances to start looking for changes, defaults to 30
    :type min_num_instances: int
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        lambda_: float = 50.0,
        min_num_instances: int = 30,
    ) -> None:
        super().__init__(min_num_instances=min_num_instances)
        self.lambda_ = lambda_

    @property
    def lambda_(self) -> float:
        """Threshold property.

        :return: lambda to use
        :rtype: float
        """
        return self._lambda

    @lambda_.setter
    def lambda_(self, value: float) -> None:
        """Threshold setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("lambda_ must be great or equal than 0.")
        self._lambda = value


class DeltaConfig:
    """Class representing a delta configuration class."""

    def __init__(
        self,
        delta: float = 0.005,
    ) -> None:
        """Init method.

        :param delta: delta value
        :type delta: float
        """
        self.delta = delta

    @property
    def delta(self) -> float:
        """Delta property.

        :return: delta to use
        :rtype: float
        """
        return self._delta

    @delta.setter
    def delta(self, value: float) -> None:
        """Delta setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if not 0.0 <= value <= 1.0:
            raise ValueError("delta must be in the range [0, 1].")
        self._delta = value


class AlphaConfig:
    """Class representing an alpha configuration class."""

    def __init__(
        self,
        alpha: float = 0.9999,
    ) -> None:
        """Init method.

        :param alpha: forgetting factor value
        :type alpha: float
        """
        self.alpha = alpha

    @property
    def alpha(self) -> float:
        """Forgetting factor property.

        :return: forgetting factor value
        :rtype: float
        """
        return self._alpha

    @alpha.setter
    def alpha(self, value: float) -> None:
        """Forgetting factor setter.

        :param value: forgetting factor value
        :type value: float
        :raises ValueError: Value error exception
        """
        if not 0.0 <= value <= 1.0:
            raise ValueError("alpha must be in the range [0, 1].")
        self._alpha = value


class BaseCUSUM(BaseChangeDetection):
    """CUSUM based algorithm class.

    :param config: configuration parameters, defaults to None
    :type config: Optional[BaseCUSUMConfig]
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]]
    """  # noqa: E501

    config_type = BaseCUSUMConfig

    def __init__(  # noqa: D107
        self,
        config: Optional[BaseCUSUMConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        super().__init__(
            config=config,
            callbacks=callbacks,
        )
        self.additional_vars = {
            "mean_error_rate": Mean(),
            "sum_": 0.0,
        }
        self._set_additional_vars_callback()

    @property
    def mean_error_rate(self) -> Mean:
        """Mean error rate property.

        :return: mean error rate to use
        :rtype: Mean
        """
        return self._additional_vars["mean_error_rate"]

    @mean_error_rate.setter
    def mean_error_rate(self, value: Mean) -> None:
        """Mean error rate setter.

        :param value: value to be set
        :type value: Mean
        """
        self._additional_vars["mean_error_rate"] = value

    @property
    def sum_(self) -> float:
        """Sum count property.

        :return: sum count value
        :rtype: float
        """
        return self._additional_vars["sum_"]

    @sum_.setter
    def sum_(self, value: float) -> None:
        """Sum count setter.

        :param value: value to be set
        :type value: float
        """
        self._additional_vars["sum_"] = value

    @abc.abstractmethod
    def _update_sum(self, error_rate: float) -> None:
        pass

    def reset(self) -> None:
        """Reset method."""
        super().reset()
        self.mean_error_rate = Mean()
        self.sum_ = 0.0

    def _update(self, value: Union[int, float], **kwargs) -> None:
        self.num_instances += 1

        self.mean_error_rate.update(value=value)
        self._update_sum(error_rate=value)

        if (
            self.num_instances >= self.config.min_num_instances  # type: ignore
            and self.sum_ > self.config.lambda_  # type: ignore
        ):
            self.drift = True
        else:
            self.drift = False



"""Data drift batch distance based detection methods' init."""

from .bhattacharyya_distance import BhattacharyyaDistance
from .emd import EMD
from .energy_distance import EnergyDistance
from .hellinger_distance import HellingerDistance
from .hi_normalized_complement import HINormalizedComplement
from .js import JS
from .kl import KL
from .mmd import MMD
from .psi import PSI

__all__ = [
    "BhattacharyyaDistance",
    "EMD",
    "EnergyDistance",
    "HellingerDistance",
    "HINormalizedComplement",
    "JS",
    "KL",
    "PSI",
    "MMD",
]



"""Prequential error using fading factor metric module."""

from typing import Optional, Union

from frouros.metrics.base import BaseMetric


class PrequentialError(BaseMetric):
    """Prequential error [dawid1984present]_ using fading factor [gama2009issues]_ metric.

    :param alpha: fading factor value, defaults to 1.0
    :type alpha: Union[int, float]
    :param name: name value, defaults to None. If None, the name will be set to `PrequentialError`.
    :type name: Optional[str]

    :References:

    .. [dawid1984present] Dawid, A. Philip.
        "Present position and potential developments:
        Some personal views statistical theory the prequential approach."
        Journal of the Royal Statistical Society:
        Series A (General) 147.2 (1984): 278-290.
    .. [gama2009issues] Gama, Joao, Raquel Sebastiao, and Pedro Pereira Rodrigues.
        "Issues in evaluation of stream learning algorithms."
        Proceedings of the 15th ACM SIGKDD international conference on Knowledge
        discovery and data mining. 2009.

    :Example:

    >>> from frouros.metrics import PrequentialError
    >>> metric = PrequentialError(alpha=0.9)
    >>> X = [1, 1, 0, 1, 0, 0]
    >>> Y = [1, 0, 0, 0, 1, 1]
    >>> for i, (X_sample, Y_sample) in enumerate(zip(X, Y)):
    ...     error_value = 1 - (X_sample == Y_sample)
    ...     prequential_error = metric(error_value=error_value)
    ...     print(f"Metric={prequential_error:.5f} at step {i}")
    Metric=0.00000 at step 0
    Metric=0.52632 at step 1
    Metric=0.33210 at step 2
    Metric=0.52632 at step 3
    Metric=0.64199 at step 4
    Metric=0.71839 at step 5
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        alpha: Union[int, float] = 1.0,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(name=name)
        self.alpha = alpha
        self.cumulative_error = 0.0
        self.cumulative_instances = 0.0
        self.num_instances = 0

    @property
    def alpha(self) -> Union[int, float]:
        """Fading factor property.

        :return: fading factor value
        :rtype: Union[int, float]
        """
        return self._alpha

    @alpha.setter
    def alpha(self, value: Union[int, float]) -> None:
        """Fading factor setter.

        :param value: value to be set
        :type value: Union[int, float]
        """
        if not isinstance(value, (int, float)):
            raise TypeError("value must be of type int or float.")
        if not 0.0 < value <= 1.0:
            raise ValueError("value must be in the range (0, 1].")
        self._alpha = value

    @property
    def cumulative_instances(self) -> Union[int, float]:
        """Cumulative instances' property.

        :return: fading factor value
        :rtype: Union[int, float]
        """
        return self._cumulative_instances

    @cumulative_instances.setter
    def cumulative_instances(self, value: Union[int, float]) -> None:
        """Cumulative instances' setter.

        :param value: value to be set
        :type value: Union[int, float]
        """
        if not isinstance(value, (int, float)):
            raise TypeError("value must be of type int or float.")
        self._cumulative_instances = value

    @property
    def cumulative_fading_error(self) -> Union[int, float]:
        """Cumulative fading error property.

        :return: cumulative facing error value
        :rtype: Union[int, float]
        """
        return self.cumulative_error / self.cumulative_instances

    def __call__(
        self,
        error_value: float,
    ) -> Union[int, float]:
        """__call__ method that updates the prequential error using fading factor.

        :param error_value error value
        :type error_value: float
        :return: cumulative facing error
        :rtype: Union[int, float]
        """
        self.cumulative_error = self.cumulative_error * self.alpha + error_value
        self.cumulative_instances = self.cumulative_instances * self.alpha + 1
        return self.cumulative_fading_error

    def reset(self) -> None:
        """Reset method."""
        self.cumulative_error = 0.0
        self.cumulative_instances = 0.0
        self.num_instances = 0

    def __repr__(self) -> str:
        """Repr method.

        :return: repr value
        :rtype: str
        """
        return f"{super().__repr__()[:-1]}, alpha={self.alpha})"



"""Test callback module."""

import numpy as np  # type: ignore
import pytest  # type: ignore

from frouros.callbacks.batch import (
    PermutationTestDistanceBased,
    ResetStatisticalTest,
)
from frouros.callbacks.streaming import (
    HistoryConceptDrift,
)
from frouros.detectors.concept_drift import (
    ADWIN,
    CUSUM,
    DDM,
    ECDDWT,
    EDDM,
    GeometricMovingAverage,
    HDDMA,
    HDDMW,
    KSWIN,
    PageHinkley,
    RDDM,
    STEPD,
)
from frouros.detectors.concept_drift.base import BaseConceptDrift
from frouros.detectors.data_drift.batch import (
    AndersonDarlingTest,
    BWSTest,
    BhattacharyyaDistance,
    CVMTest,
    EMD,
    EnergyDistance,
    HellingerDistance,
    HINormalizedComplement,
    JS,
    KL,
    KSTest,
    KuiperTest,
    MannWhitneyUTest,
    MMD,
    PSI,
    WelchTTest,
)
from frouros.detectors.data_drift.batch.base import BaseDataDriftBatch


@pytest.mark.parametrize(
    "detector_class, expected_distance, expected_p_value",
    [
        (BhattacharyyaDistance, 0.55516059, 0.0),
        (EMD, 3.85346006, 0.0),
        (EnergyDistance, 2.11059982, 0.0),
        (HellingerDistance, 0.74509099, 0.0),
        (HINormalizedComplement, 0.78, 0.0),
        (JS, 0.67010107, 0.0),
        (KL, np.inf, 0.06),
        (MMD, 0.69509004, 0.0),
        (PSI, 461.20379435, 0.0),
    ],
)
def test_batch_permutation_test_data_univariate_different_distribution(
    X_ref_univariate: np.ndarray,  # noqa: N803
    X_test_univariate: np.ndarray,
    detector_class: BaseDataDriftBatch,
    expected_distance: float,
    expected_p_value: float,
) -> None:
    """Test batch permutation test on data drift callback.

    :param X_ref_univariate: reference univariate data
    :type X_ref_univariate: numpy.ndarray
    :param X_test_univariate: test univariate data
    :type X_test_univariate: numpy.ndarray
    :param detector_class: detector distance
    :type detector_class: BaseDataDriftBatch
    :param expected_distance: expected distance value
    :type expected_distance: float
    :param expected_p_value: expected p-value value
    :type expected_p_value: float
    """
    np.random.seed(seed=31)

    permutation_test_name = "permutation_test"
    detector = detector_class(  # type: ignore
        callbacks=[
            PermutationTestDistanceBased(
                num_permutations=100,
                method="estimate",
                random_state=31,
                num_jobs=-1,
                name=permutation_test_name,
            )
        ]
    )
    _ = detector.fit(X=X_ref_univariate)
    distance, callback_logs = detector.compare(X=X_test_univariate)

    assert np.isclose(distance, expected_distance)
    assert np.isclose(
        callback_logs[permutation_test_name]["p_value"],
        expected_p_value,
    )


@pytest.mark.parametrize(
    "method, expected_p_value",
    [
        ("auto", 0.009900490107343236),
        ("conservative", 0.009900990099009901),
        ("exact", 0.009900490107343236),
        ("approximate", 0.009900990098759907),
        ("estimate", 0.0),
    ],
)
def test_batch_permutation_test_method(
    X_ref_univariate: np.ndarray,  # noqa: N803
    X_test_univariate: np.ndarray,
    method: str,
    expected_p_value: float,
) -> None:
    """Test batch permutation test on data drift callback using method.

    :param X_ref_univariate: reference univariate data
    :type X_ref_univariate: numpy.ndarray
    :param X_test_univariate: test univariate data
    :type X_test_univariate: numpy.ndarray
    :param method: method
    :type method: str
    :param expected_p_value: expected p-value value
    :type expected_p_value: float
    """
    np.random.seed(seed=31)

    permutation_test_name = "permutation_test"
    detector = MMD(  # type: ignore
        callbacks=[
            PermutationTestDistanceBased(
                num_permutations=100,
                method=method,
                random_state=31,
                num_jobs=-1,
                name=permutation_test_name,
            )
        ]
    )
    _ = detector.fit(X=X_ref_univariate)
    _, callback_logs = detector.compare(X=X_test_univariate)

    assert np.isclose(
        callback_logs[permutation_test_name]["p_value"],
        expected_p_value,
    )


@pytest.mark.parametrize(
    "detector_class",
    [
        AndersonDarlingTest,
        BWSTest,
        CVMTest,
        KSTest,
        KuiperTest,
        MannWhitneyUTest,
        WelchTTest,
    ],
)
def test_batch_reset_on_statistical_test_data_drift(
    X_ref_univariate,  # noqa: N803
    X_test_univariate,
    detector_class: BaseDataDriftBatch,
    mocker,
) -> None:
    """Test batch reset on statistical test data drift callback.

    :param X_ref_univariate: reference univariate data
    :type X_ref_univariate: numpy.ndarray
    :param X_test_univariate: test univariate data
    :type X_test_univariate: numpy.ndarray
    :param detector_class: detector distance
    :type detector_class: BaseDataDriftBatch
    :param mocker: mocker
    :type mocker: pytest_mock.mocker
    """
    mocker.patch("frouros.detectors.data_drift.batch.base.BaseDataDriftBatch.reset")

    detector = detector_class(  # type: ignore
        callbacks=[
            ResetStatisticalTest(
                alpha=0.01,
            ),
        ],
    )
    _ = detector.fit(X=X_ref_univariate)
    _ = detector.compare(X=X_test_univariate)
    detector.reset.assert_called_once()  # type: ignore # pylint: disable=no-member


@pytest.mark.parametrize(
    "detector_class",
    [
        ADWIN,
        CUSUM,
        DDM,
        ECDDWT,
        EDDM,
        GeometricMovingAverage,
        HDDMA,
        HDDMW,
        KSWIN,
        PageHinkley,
        RDDM,
        STEPD,
    ],
)
def test_streaming_history_on_concept_drift(
    model_errors: list[int],
    detector_class: BaseConceptDrift,
):
    """Test streaming history on concept drift callback.

    :param model_errors: model errors
    :type model_errors: list[int]
    :param detector_class: concept drift detector
    :type detector_class: BaseConceptDrift
    """
    name = "history"
    detector = detector_class(callbacks=HistoryConceptDrift(name=name))  # type: ignore

    for error in model_errors:
        history = detector.update(value=error)
        status = detector.status
        if status["drift"]:
            assert history[name]["drift"][-1]
            assert not any(history[name]["drift"][:-1])
            break



"""Concept drift streaming detection methods init."""
# FIXME: Remove pylint disable if batch methods are added
# pylint: skip-file
from .change_detection import (
    BOCD,
    BOCDConfig,
    CUSUM,
    CUSUMConfig,
    GeometricMovingAverage,
    GeometricMovingAverageConfig,
    PageHinkley,
    PageHinkleyConfig,
)
from .statistical_process_control import (
    DDM,
    DDMConfig,
    ECDDWT,
    ECDDWTConfig,
    EDDM,
    EDDMConfig,
    HDDMA,
    HDDMAConfig,
    HDDMW,
    HDDMWConfig,
    RDDM,
    RDDMConfig,
)
from .window_based import (
    ADWIN,
    ADWINConfig,
    KSWIN,
    KSWINConfig,
    STEPD,
    STEPDConfig,
)

__all__ = [
    "ADWIN",
    "ADWINConfig",
    "BOCD",
    "BOCDConfig",
    "CUSUM",
    "CUSUMConfig",
    "DDM",
    "DDMConfig",
    "ECDDWT",
    "ECDDWTConfig",
    "EDDM",
    "EDDMConfig",
    "GeometricMovingAverage",
    "GeometricMovingAverageConfig",
    "HDDMA",
    "HDDMAConfig",
    "HDDMW",
    "HDDMWConfig",
    "KSWIN",
    "KSWINConfig",
    "PageHinkley",
    "PageHinkleyConfig",
    "RDDM",
    "RDDMConfig",
    "STEPD",
    "STEPDConfig",
]



"""ChiSquareTest (Chi-square test) module."""

import collections
import typing
from typing import Optional, Set, Tuple, Union

import numpy as np  # type: ignore
from scipy.stats import chi2_contingency  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.base import CategoricalData, UnivariateData
from frouros.detectors.data_drift.batch.statistical_test.base import (  # type: ignore
    BaseStatisticalTest,
    StatisticalResult,
)


class ChiSquareTest(BaseStatisticalTest):
    """ChiSquareTest (Chi-square test) [pearson1900x]_ detector.

    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]]

    :Note:
    - Passing additional arguments to `scipy.stats.chi2_contingency <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.chi2_contingency.html>`__ can be done using :func:`compare` kwargs.

    :References:

    .. [pearson1900x] Pearson, Karl.
        "X. On the criterion that a given system of deviations from the probable in the
        case of a correlated system of variables is such that it can be reasonably
        supposed to have arisen from random sampling."
        The London, Edinburgh, and Dublin Philosophical Magazine and Journal of
        Science 50.302 (1900): 157-175.

    :Example:

    >>> from frouros.detectors.data_drift import ChiSquareTest
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> X = np.random.choice(a=[0, 1], size=100, p=[0.5, 0.5])
    >>> Y = np.random.choice(a=[0, 1], size=100, p=[0.8, 0.2])
    >>> detector = ChiSquareTest()
    >>> _ = detector.fit(X=X)
    >>> detector.compare(X=Y)[0]
    StatisticalResult(statistic=9.81474665685192, p_value=0.0017311812135839511)
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]] = None,
    ) -> None:
        super().__init__(
            data_type=CategoricalData(),
            statistical_type=UnivariateData(),
            callbacks=callbacks,
        )

    @staticmethod
    def _statistical_test(
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,
        **kwargs,
    ) -> StatisticalResult:
        f_exp, f_obs = ChiSquareTest._calculate_frequencies(
            X_ref=X_ref,
            X=X,
        )
        statistic, p_value, _, _ = chi2_contingency(
            observed=np.array([f_obs, f_exp]),
            **kwargs,
        )

        test = StatisticalResult(
            statistic=statistic,
            p_value=p_value,
        )
        return test

    @staticmethod
    @typing.no_type_check  # FIXME: X_ref_counter and X_counter cause mypy errors  # pylint: disable=fixme  # noqa: E501
    def _calculate_frequencies(
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,
    ) -> Tuple[list[int], list[int]]:
        X_ref_counter, X_counter = [  # noqa: N806
            *map(collections.Counter, [X_ref, X])  # noqa: N806
        ]
        possible_values: Set[str] = set(
            [*X_ref_counter.keys()] + [*X_counter.keys()]
        )  # noqa: N806
        f_exp, f_obs = {}, {}
        for value in possible_values:
            f_exp[value] = X_ref_counter.get(value, 0)  # noqa: N806
            f_obs[value] = X_counter.get(value, 0)  # noqa: N806
        f_exp_values, f_obs_values = [*map(list, [f_exp.values(), f_obs.values()])]
        return f_exp_values, f_obs_values



"""Synthetic datasets module."""

from typing import Tuple, Iterator, Optional

import numpy as np  # type: ignore

from frouros.datasets.base import BaseDatasetGenerator
from frouros.datasets.exceptions import InvalidBlockError


class SEA(BaseDatasetGenerator):
    """SEA generator [street2001streaming]_.

    :param seed: seed value, defaults to None
    :type seed: Optional[int]

    :References:

    .. [street2001streaming] Street, W. Nick, and YongSeog Kim.
        "A streaming ensemble algorithm (SEA) for large-scale classification."
        Proceedings of the seventh ACM SIGKDD international conference on Knowledge
        discovery and data mining. 2001.

    :Example:

    >>> from frouros.datasets.synthetic import SEA
    >>> sea = SEA(seed=31)
    >>> dataset = sea.generate_dataset(block=1, noise=0.1, num_samples=5)
    >>> for X, y in dataset:
    ...     print(X, y)
    [2.86053822 9.58105567 7.70312932] 0
    [2.08165462 1.36917049 9.08373802] 0
    [8.36483632 1.12172604 8.3489916 ] 0
    [2.44680795 1.36231348 7.22094455] 1
    [1.28477715 2.20364007 5.19211202] 1
    """

    def __init__(  # noqa: D107
        self,
        seed: Optional[int] = None,
    ) -> None:
        super().__init__(
            seed=seed,
        )
        self._block_map = {1: 8.0, 2: 9.0, 3: 7.0, 4: 9.5}

    @staticmethod
    def _generate_sample(threshold: float, noise: float) -> Tuple[np.ndarray, int]:
        X = np.random.uniform(low=0.0, high=10.0, size=(3,))  # noqa: N806
        if np.random.random() < noise:
            y = np.random.randint(2)
        else:
            y = 1 if X[0] + X[1] <= threshold else 0
        return X, y

    def generate_dataset(
        self, block: int, noise: float = 0.1, num_samples: int = 12500
    ) -> Iterator[Tuple[np.ndarray, int]]:
        """Generate dataset.

        :param block: block to generate samples from, must be 1, 2, 3 or 4
        :type block: int
        :param noise: ratio of samples with a noisy class, defaults to 0.1
        :type noise: float
        :param num_samples: number of samples to generate, defaults to 12500
        :type num_samples: int
        :return: generator with the samples
        :rtype: Iterator[Tuple[np.ndarray, int]]
        """
        try:
            threshold = self._block_map[block]
        except KeyError as e:
            raise InvalidBlockError("block must be 1, 2, 3 or 4.") from e
        if num_samples < 1:
            raise ValueError("num_samples must be greater than 0.")
        if not 0 <= noise <= 1:
            raise ValueError("noise must be in the range [0, 1].")
        dataset = (
            self._generate_sample(threshold=threshold, noise=noise)
            for _ in range(num_samples)
        )
        return dataset


class Dummy(BaseDatasetGenerator):
    """Dummy generator class."""

    @staticmethod
    def _generate_sample(class_: int) -> Tuple[np.ndarray, int]:
        X = np.random.uniform(low=0.0, high=10.0, size=(2,))  # noqa: N806
        y = class_ if X[0] + X[1] < 10.0 else 1 - class_
        return X, y

    def generate_dataset(
        self, class_: int = 1, num_samples: int = 12500
    ) -> Iterator[Tuple[np.ndarray, int]]:
        """Generate dataset.

        :param class_: class value
        :type class_: int
        :param num_samples: number of samples to generate
        :type num_samples: int
        :return: generator with the samples
        :rtype: Iterator[Tuple[np.ndarray, int]]
        """
        if class_ not in [1, 0]:
            raise ValueError("class_ must be 1 or 0.")
        if num_samples < 1:
            raise ValueError("num_samples must be greater than 0.")
        dataset = (self._generate_sample(class_=class_) for _ in range(num_samples))
        return dataset



"""RDDM (Reactive Drift detection method) module."""

from typing import Optional, Union

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.concept_drift.streaming.statistical_process_control.base import (
    BaseSPCConfig,
    BaseSPCError,
)
from frouros.utils.data_structures import CircularQueue
from frouros.utils.stats import Mean


class RDDMConfig(BaseSPCConfig):
    """RDDM (Reactive Drift detection method) [barros2017rddm]_ configuration.

    :param warning_level: warning level factor, defaults to 1.773
    :type warning_level: float
    :param drift_level: drift level factor, defaults to 2.258
    :type drift_level: float
    :param max_concept_size: maximum size of a concept, defaults to 40000
    :type max_concept_size: int
    :param min_concept_size: reduced size of a concept, defaults to 7000
    :type min_concept_size: int
    :param max_num_instances_warning: maximum number of instances at warning level, defaults to 1400
    :type max_num_instances_warning: int
    :param min_num_instances: minimum numbers of instances to start looking for changes, defaults to 129
    :type min_num_instances: int

    :References:

    .. [barros2017rddm] Barros, Roberto SM, et al.
        "RDDM: Reactive drift detection method."
        Expert Systems with Applications 90 (2017): 344-355.
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        warning_level: float = 1.773,
        drift_level: float = 2.258,
        max_concept_size: int = 40000,
        min_concept_size: int = 7000,
        max_num_instances_warning: int = 1400,
        min_num_instances: int = 129,
    ) -> None:
        super().__init__(
            drift_level=drift_level,
            warning_level=warning_level,
            min_num_instances=min_num_instances,
        )
        self.max_concept_size = max_concept_size
        self.min_concept_size = min_concept_size
        self.max_num_instances_warning = max_num_instances_warning

    @property
    def max_concept_size(self) -> int:
        """Maximum size of a concept property.

        :return: maximum size of a concept
        :rtype: int
        """
        return self._max_concept_size

    @max_concept_size.setter
    def max_concept_size(self, value: int) -> None:
        """Maximum size of a concept setter.

        :param value: value to be set
        :type value: int
        """
        self._max_concept_size = value

    @property
    def min_concept_size(self) -> int:
        """Minimum size of a concept property.

        :return: minimum size of a concept
        :rtype: int
        """
        return self._min_concept_size

    @min_concept_size.setter
    def min_concept_size(self, value: int) -> None:
        """Minimum size of a concept setter.

        :param value: value to be set
        :type value: int
        """
        self._min_concept_size = value

    @property
    def max_num_instances_warning(self) -> int:
        """Maximum number of instances at warning level property.

        :return: maximum number of instances at warning level
        :rtype: int
        """
        return self._max_num_instances_warning

    @max_num_instances_warning.setter
    def max_num_instances_warning(self, value: int) -> None:
        """Maximum number of instances at warning level setter.

        :param value: value to be set
        :type value: int
        """
        self._max_num_instances_warning = value


class RDDM(BaseSPCError):
    """RDDM (Reactive Drift detection method) [barros2017rddm]_ detector.

    :param config: configuration object of the detector, defaults to None. If None, the default configuration of :class:`RDDMConfig` is used.
    :type config: Optional[RDDMConfig]
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]]

    :Note:
    :func:`update` method expects to receive a value of 0 if the instance is correctly classified (no error) and 1 otherwise (error).

    :References:

    .. [barros2017rddm] Barros, Roberto SM, et al.
        "RDDM: Reactive drift detection method."
        Expert Systems with Applications 90 (2017): 344-355.

    :Example:

    >>> from frouros.detectors.concept_drift import RDDM
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> dist_a = np.random.binomial(n=1, p=0.6, size=1000)
    >>> dist_b = np.random.binomial(n=1, p=0.8, size=1000)
    >>> stream = np.concatenate((dist_a, dist_b))
    >>> detector = RDDM()
    >>> warning_flag = False
    >>> for i, value in enumerate(stream):
    ...     _ = detector.update(value=value)
    ...     if detector.drift:
    ...         print(f"Change detected at step {i}")
    ...         break
    ...     if not warning_flag and detector.warning:
    ...         print(f"Warning detected at step {i}")
    ...         warning_flag = True
    Warning detected at step 1036
    Change detected at step 1066
    """  # noqa: E501  # pylint: disable=line-too-long

    config_type = RDDMConfig  # type: ignore

    def __init__(  # noqa: D107
        self,
        config: Optional[RDDMConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        super().__init__(
            config=config,
            callbacks=callbacks,
        )
        self.additional_vars = {
            "num_warnings": 0,
            "rddm_drift": False,
            "predictions": CircularQueue(
                max_len=self.config.min_concept_size  # type: ignore
            ),
            **self.additional_vars,  # type: ignore
        }
        self._set_additional_vars_callback()

    @property
    def num_warnings(self) -> int:
        """Number of warnings property.

        :return: number of warnings
        :rtype: int
        """
        return self._additional_vars["num_warnings"]

    @num_warnings.setter
    def num_warnings(self, value: int) -> None:
        """Number of warnings setter.

        :param value: value to be set
        :type value: int
        """
        self._additional_vars["num_warnings"] = value

    @property
    def rddm_drift(self) -> bool:
        """Rddm drift property.

        :return: rddmi drift value
        :rtype: bool
        """
        return self._additional_vars["rddm_drift"]

    @rddm_drift.setter
    def rddm_drift(self, value: bool) -> None:
        """Rddm drift setter.

        :param value: value to be set
        :type value: bool
        """
        self._additional_vars["rddm_drift"] = value

    @property
    def predictions(self) -> CircularQueue:
        """Predictions circular queue property.

        :return: predictions circular queue
        :rtype: CircularQueue
        """
        return self._additional_vars["predictions"]

    @predictions.setter
    def predictions(self, value: CircularQueue) -> None:
        """Predictions circular queue setter.

        :param value: value to be set
        :type value: CircularQueue
        :raises TypeError: Type error exception
        """
        if not isinstance(value, CircularQueue):
            raise TypeError("value must be of type CircularQueue.")
        self._additional_vars["predictions"] = value

    def reset(self) -> None:
        """Reset method."""
        super().reset()
        self.rddm_drift = False

    def _update(  # pylint: disable=too-many-branches
        self,
        value: Union[int, float],
        **kwargs,
    ) -> None:
        self.num_instances += 1

        if self.rddm_drift:
            self._rdd_drift_case()

        self.predictions.enqueue(value=value)
        self.error_rate.update(value=value)

        if self.num_instances >= self.config.min_num_instances:
            error_rate_plus_std, std = self._calculate_error_rate_plus_std()

            self._update_min_values(error_rate_plus_std=error_rate_plus_std, std=std)

            drift_flag = self._check_threshold(
                error_rate_plus_std=error_rate_plus_std,
                min_error_rate=self.min_error_rate,
                min_std=self.min_std,
                level=self.config.drift_level,  # type: ignore
            )

            if drift_flag:
                # Out-of-Control
                self.rddm_drift = True
                self.drift = True
                self.warning = False
                if self.num_warnings == 0:
                    self.predictions.maintain_last_element()
            else:
                warning_flag = self._check_threshold(
                    error_rate_plus_std=error_rate_plus_std,
                    min_error_rate=self.min_error_rate,
                    min_std=self.min_std,
                    level=self.config.warning_level,  # type: ignore
                )
                if warning_flag:
                    if (
                        self.num_warnings
                        >= self.config.max_num_instances_warning  # type: ignore
                    ):
                        self.rddm_drift = True
                        self.drift = True
                        self.predictions.maintain_last_element()
                    else:
                        # Warning
                        self.warning = True
                        self.num_warnings += 1
                        self.drift = False
                else:
                    # In-Control
                    self.drift = False
                    self.warning = False
                    self.num_warnings = 0
                if (
                    self.num_instances >= self.config.max_concept_size  # type: ignore
                    and not self.warning
                ):
                    self.rddm_drift = True
        else:
            self.drift, self.warning = False, False

    def _rdd_drift_case(self) -> None:
        self._reset_stats()
        pos = self.predictions.first
        for _ in range(self.predictions.count):
            self.num_instances += 1
            self.error_rate.update(value=self.predictions[pos])
            error_rate_plus_std, std = self._calculate_error_rate_plus_std()

            if (
                self.drift
                and self.num_instances >= self.config.min_num_instances
                and error_rate_plus_std < self.min_error_rate_plus_std
            ):
                self.min_error_rate = self.error_rate.mean
                self.min_std = std

            pos = (pos + 1) % self.config.min_concept_size  # type: ignore
        self.rddm_drift = False
        self.drift = False

    def _reset_stats(self):
        self.error_rate = Mean()
        self.min_error_rate = float("inf")
        self.min_std = float("inf")
        self.num_warnings = 0
        self.num_instances = 0
        self.rddm_drift = False



"""BaseDatasetDownload exception module."""


class DownloadError(Exception):
    """Download exception."""


class InvalidBlockError(Exception):
    """Invalid block exception."""


class InvalidFilePathError(Exception):
    """Invalid file path exception."""


class InvalidURLError(Exception):
    """Invalid URL exception."""


class RequestFileError(Exception):
    """Request file exception."""


class ReadFileError(Exception):
    """Read file exception."""



"""ADWIN (ADaptive WINdowing) module."""

from collections import deque
from typing import Optional, Union

import numpy as np  # type: ignore

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.concept_drift.streaming.window_based.base import (
    BaseWindowConfig,
    BaseWindow,
)


class Bucket:
    """Class representing a bucket."""

    def __init__(self, m: int) -> None:
        """Init method.

        :param m: controls the amount of memory used and
        the closeness of the cutpoints checked
        :type m: int
        """
        self.array_size = m + 1
        self.total = np.zeros(self.array_size)
        self.variance = np.zeros(self.array_size)
        self.idx = 0

    @property
    def array_size(self) -> int:
        """Array size property.

        :return: array size
        :rtype: int
        """
        return self._array_size

    @array_size.setter
    def array_size(self, value: int) -> None:
        """Array size setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 2:
            raise ValueError("array_size value must be greater than 1.")
        self._array_size = value

    @property
    def total(self) -> np.ndarray:
        """Total array property.

        :return: total array
        :rtype: numpy.ndarray
        """
        return self._total

    @total.setter
    def total(self, value: np.ndarray) -> None:
        """Total array setter.

        :param value: value to be set
        :type value: numpy.ndarray
        :raises TypeError: Type error exception
        """
        if not isinstance(value, np.ndarray):
            raise TypeError("value must be of type numpy.ndarray.")
        self._total = value

    @property
    def variance(self) -> np.ndarray:
        """Variance array property.

        :return: variance array
        :rtype: numpy.ndarray
        """
        return self._variance

    @variance.setter
    def variance(self, value: np.ndarray) -> None:
        """Variance array setter.

        :param value: value to be set
        :type value: numpy.ndarray
        :raises TypeError: Type error exception
        """
        if not isinstance(value, np.ndarray):
            raise TypeError("value must be of type numpy.ndarray.")
        self._variance = value

    @property
    def idx(self) -> int:
        """Current index property.

        :return: current index
        :rtype: int
        """
        return self._idx

    @idx.setter
    def idx(self, value: int):
        """Current index setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("idx value must be greater or equal than 0.")
        self._idx = value

    def reset(self) -> None:
        """Reset bucket statistics."""
        self.total = np.zeros(self.array_size)
        self.variance = np.zeros(self.array_size)

    def insert_data(self, value: float, variance: float) -> None:
        """Insert data at the end.

        :param value: value to be stored
        :type value: float
        :param variance: variance to be stored
        :type variance: float
        """
        self.total[self.idx] = value
        self.variance[self.idx] = variance
        self.idx += 1

    def remove(self) -> None:
        """Remove first."""
        self.compress(num_items_deleted=1)

    def compress(self, num_items_deleted: int) -> None:
        """Compress bucket index.

        :param num_items_deleted: number of items to delete
        :type num_items_deleted: int
        """
        for i in range(num_items_deleted, self.array_size):
            idx = i - num_items_deleted
            self.total[idx] = self.total[i]
            self.variance[idx] = self.variance[i]

        idx_start = self.array_size - num_items_deleted
        # fmt: off
        self.total[idx_start:self.array_size] = 0.0
        self.variance[idx_start:self.array_size] = 0.0
        # fmt: on

        self.idx -= num_items_deleted


class ADWINConfig(BaseWindowConfig):
    """ADWIN (ADaptive WINdowing) [bifet2007learning]_ configuration.

    :param clock: clock value, default to 32
    :type clock: int
    :param delta: confidence value, default to 0.002
    :type delta: float
    :param m: controls the amount of memory used and the closeness of the cutpoints checked, default to 5
    :type m: int
    :param min_window_size: minimum numbers of instances per window to start looking for changes, default to 5
    :type min_window_size: int
    :param min_num_instances: minimum numbers of instances to start looking for changes, default to 10
    :type min_num_instances: int

    :References:

    .. [bifet2007learning] Bifet, Albert, and Ricard Gavalda.
        "Learning from time-changing data with adaptive windowing."
        Proceedings of the 2007 SIAM international conference on data mining.
        Society for Industrial and Applied Mathematics, 2007.
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        clock: int = 32,
        delta: float = 0.002,
        m: int = 5,
        min_window_size: int = 5,
        min_num_instances: int = 10,
    ) -> None:
        super().__init__(min_num_instances=min_num_instances)
        self.clock = clock
        self.delta = delta
        self.m = m
        self.min_window_size = min_window_size

    @property
    def clock(self) -> int:
        """Clock value property.

        :return: confidence interval to determine if drift is occurring
        :rtype: int
        """
        return self._clock

    @clock.setter
    def clock(self, value: int) -> None:
        """Clock value setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("clock value must be greater than 0.")
        self._clock = value

    @property
    def delta(self) -> float:
        """Delta value property.

        :return: confidence interval to determine if drift is occurring
        :rtype: float
        """
        return self._delta

    @delta.setter
    def delta(self, value: float) -> None:
        """Delta value setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if not 0 < value < 1:
            raise ValueError("delta value must be in the range (0, 1).")
        self._delta = value

    @property
    def m(self) -> int:
        """M value property.

        :return: controls the amount of memory used and the closeness
        of the cutpoints checked
        :rtype: int
        """
        return self._m

    @m.setter
    def m(self, value: int) -> None:
        """M value setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value < 1:
            raise ValueError("m value must be greater than 0.")
        self._m = value

    @property
    def min_window_size(self) -> int:
        """Minimum window size value property.

        :return: minimum window size value per each window
        :rtype: int
        """
        return self._min_window_size

    @min_window_size.setter
    def min_window_size(self, value: int) -> None:
        """Minimum window size value setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value < 1:
            raise ValueError("min_window_size value must be greater than 0.")
        self._min_window_size = value


class ADWIN(BaseWindow):
    """ADWIN (ADaptive WINdowing) [bifet2007learning]_ detector.

    :param config: configuration object of the detector, defaults to None. If None, the default configuration of :class:`ADWINConfig` is used.
    :type config: Optional[ADWINConfig]
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]]

    :References:

    .. [bifet2007learning] Bifet, Albert, and Ricard Gavalda.
        "Learning from time-changing data with adaptive windowing."
        Proceedings of the 2007 SIAM international conference on data mining.
        Society for Industrial and Applied Mathematics, 2007.

    :Example:

    >>> from frouros.detectors.concept_drift import ADWIN
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> dist_a = np.random.normal(loc=0.2, scale=0.01, size=1000)
    >>> dist_b = np.random.normal(loc=0.8, scale=0.04, size=1000)
    >>> stream = np.concatenate((dist_a, dist_b))
    >>> detector = ADWIN()
    >>> for i, value in enumerate(stream):
    ...     _ = detector.update(value=value)
    ...     if detector.drift:
    ...         print(f"Change detected at step {i}")
    ...         break
    Change detected at step 1055
    """  # noqa: E501  # pylint: disable=line-too-long

    config_type = ADWINConfig

    def __init__(  # noqa: D107
        self,
        config: Optional[ADWINConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        super().__init__(
            config=config,
            callbacks=callbacks,
        )
        num_buckets = 0
        self.additional_vars = {
            "buckets": deque([Bucket(m=self.config.m)]),  # type: ignore
            "total": 0.0,
            "variance": 0.0,
            "width": 0,
            "num_buckets": num_buckets,
            "num_max_buckets": num_buckets,
        }
        self._set_additional_vars_callback()
        self._min_instances = self.config.min_num_instances + 1

    @property
    def buckets(self) -> deque:
        """Buckets queue property.

        :return: buckets queue
        :rtype: deque
        """
        return self._additional_vars["buckets"]

    @buckets.setter
    def buckets(self, value: deque):
        """Buckets queue setter.

        :param value: value to be set
        :type value: int
        :raises TypeError: Type error exception
        """
        if not isinstance(value, deque):
            raise TypeError("value must be of type deque.")
        self._additional_vars["buckets"] = value

    @property
    def total(self) -> float:
        """Total value property.

        :return: total value
        :rtype: float
        """
        return self._additional_vars["total"]

    @total.setter
    def total(self, value: float) -> None:
        """Total value setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value < 0.0:
            raise ValueError("total value must be greater or equal than 0.0.")
        self._additional_vars["total"] = value

    @property
    def variance(self) -> float:
        """Variance value property.

        :return: variance value
        :rtype: float
        """
        return self._additional_vars["variance"]

    @variance.setter
    def variance(self, value: float) -> None:
        """Variance value setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        # FIXME: Workaround to avoid precision problems  # pylint: disable=fixme
        # if value < 0.0:
        #     raise ValueError("variance value must be greater or equal than 0.0.")
        self._additional_vars["variance"] = value

    @property
    def variance_window(self) -> float:
        """Variance in window value property.

        :return: variance in window value
        :rtype: float
        """
        return self.variance / self.width

    @property
    def width(self) -> int:
        """Width value property.

        :return: width value
        :rtype: int
        """
        return self._additional_vars["width"]

    @width.setter
    def width(self, value: int) -> None:
        """Width value setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("width value must be greater or equal than 0.")
        self._additional_vars["width"] = value

    @property
    def num_buckets(self) -> int:
        """Number of buckets property.

        :return: number of buckets
        :rtype: int
        """
        return self._additional_vars["num_buckets"]

    @num_buckets.setter
    def num_buckets(self, value: int) -> None:
        """Number of buckets setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("num_buckets value must be greater or equal than 0.")
        self._additional_vars["num_buckets"] = value

    @property
    def num_max_buckets(self) -> int:
        """Maximum number of buckets property.

        :return: maximum number of buckets
        :rtype: int
        """
        return self._additional_vars["num_max_buckets"]

    @num_max_buckets.setter
    def num_max_buckets(self, value: int) -> None:
        """Maximum number of buckets setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("num_max_buckets value must be greater or equal than 0.")
        self._additional_vars["num_max_buckets"] = value

    def _insert_bucket(self, value: float) -> None:
        self._insert_bucket_data(variance=0.0, value=value, bucket=self.buckets[0])
        self.width += 1
        incremental_variance = (
            (self.width - 1)
            * (value - self.total / (self.width - 1))
            * (value - self.total / (self.width - 1))
            / self.width
            if self.width > 1
            else 0.0
        )
        self.variance += incremental_variance
        self.total += value
        self._compress_buckets()

    def _insert_bucket_data(
        self, value: float, variance: float, bucket: Bucket
    ) -> None:
        bucket.insert_data(value=value, variance=variance)
        self.num_buckets += 1
        if self.num_buckets > self.num_max_buckets:
            self.num_max_buckets = self.num_buckets

    @staticmethod
    def _bucket_size(index: int) -> int:
        return np.power(2, index)

    def _delete_bucket(self) -> int:
        bucket = self.buckets[-1]
        bucket_size = self._bucket_size(index=len(self.buckets) - 1)
        self.width -= bucket_size
        self.total -= bucket.total[0]
        bucket_mean = bucket.total[0] / bucket_size
        window_mean = self.total / self.width
        incremental_variance = bucket.variance[0] + bucket_size * self.width * (
            bucket_mean - window_mean
        ) * (bucket_mean - window_mean) / (bucket_size + self.width)
        self.variance -= incremental_variance

        bucket.remove()
        self.num_buckets -= 1
        if bucket.idx == 0:
            self.buckets.pop()

        return bucket_size

    def _compress_buckets(self) -> None:
        bucket = self.buckets[0]
        idx = 0
        while bucket is not None:
            idx_next = idx + 1
            if bucket.idx == bucket.array_size:
                try:
                    next_bucket = self.buckets[idx_next]
                except IndexError:
                    self.buckets.append(Bucket(m=self.config.m))  # type: ignore
                    next_bucket = self.buckets[-1]
                bucket_size = self._bucket_size(index=idx)
                bucket_1_mean, bucket_2_mean = bucket.total[0:2] / bucket_size
                incremental_variance = (
                    bucket_size
                    * bucket_size
                    * (bucket_1_mean - bucket_2_mean)
                    * (bucket_1_mean - bucket_2_mean)
                    / (bucket_size * 2)
                )
                total = np.sum(bucket.total[0:2])
                variance = np.sum(bucket.variance[0:2]) + incremental_variance
                next_bucket.insert_data(value=total, variance=variance)
                self.num_buckets += 1
                bucket.compress(num_items_deleted=2)

                if next_bucket.idx <= self.config.m:  # type: ignore
                    break
            else:
                break

            try:
                bucket = self.buckets[idx_next]
            except IndexError:
                bucket = None
            idx += 1

    def _calculate_threshold(self, w0_instances: int, w1_instances: int) -> float:
        delta_prime = np.log(2 * np.log(self.width) / self.config.delta)  # type: ignore
        # Has highlighted in river library, the use of the reciprocal
        # of m allows to avoid extra divisions
        min_window_size = self.config.min_window_size + 1  # type: ignore
        m_reciprocal = 1 / (w0_instances - min_window_size) + 1 / (
            w1_instances - min_window_size
        )
        epsilon = (
            np.sqrt(2 * m_reciprocal * self.variance_window * delta_prime)
            + 2 / 3 * delta_prime * m_reciprocal
        )
        return epsilon

    def _update(self, value: Union[int, float], **kwargs) -> None:
        # pylint: disable=too-many-locals, too-many-nested-blocks
        # NOTE: Refactor function
        self.num_instances += 1
        self._insert_bucket(value=value)

        if (
            self.num_instances % self.config.clock == 0  # type: ignore
            and self.width > self.config.min_num_instances  # type: ignore
        ):
            flag_reduce_width = True

            while flag_reduce_width:
                flag_reduce_width = False
                flag_exit = False
                w0_instances = 0
                w1_instances = self.width
                w0_total = 0
                w1_total = self.total

                for i in range(len(self.buckets) - 1, -1, -1):
                    if flag_exit:
                        break
                    bucket = self.buckets[i]
                    for j in range(bucket.idx - 1):
                        bucket_size = self._bucket_size(index=i)

                        w0_instances += bucket_size
                        w1_instances -= bucket_size
                        w0_total += bucket.total[j]
                        w1_total -= bucket.total[j]

                        if i == 0 and j == bucket.idx - 1:
                            flag_exit = True
                            break

                        if (
                            w1_instances > self.config.min_window_size  # type: ignore
                            and (
                                w0_instances
                                > self.config.min_window_size  # type: ignore
                            )
                        ):
                            w0_mean = w0_total / w0_instances
                            w1_mean = w1_total / w1_instances
                            threshold = self._calculate_threshold(
                                w0_instances=w0_instances, w1_instances=w1_instances
                            )
                            if np.abs(w0_mean - w1_mean) > threshold:
                                # Drift detected
                                flag_reduce_width = True
                                self.drift = True
                                if self.width > 0:
                                    w0_instances -= self._delete_bucket()
                                    flag_exit = True
                                    break

    def reset(self) -> None:
        """Reset method."""
        super().reset()
        self.buckets = deque([Bucket(m=self.config.m)])  # type: ignore
        self.total = 0.0
        self.variance = 0.0
        self.width = 0
        self.num_buckets = 0
        self.num_max_buckets = self.num_buckets



"""Permutation test batch callback module."""

import multiprocessing
from typing import Any, Callable, Optional, Tuple

import numpy as np  # type: ignore
from scipy.integrate import quad  # type: ignore
from scipy.stats import binom  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.utils.stats import permutation

MAX_NUM_PERM: int = 1000000  # maximum number of permutations


class PermutationTestDistanceBased(BaseCallbackBatch):
    """Permutation test callback class that can be applied to :mod:`data_drift.batch.distance_based <frouros.detectors.data_drift.batch.distance_based>` detectors.

    :param num_permutations: number of permutations to obtain the p-value
    :type num_permutations: int
    :param total_num_permutations: total number of permutations to obtain the p-value, defaults to None. If None, the total number of permutations will be set to the maximum number of permutations, the minimum between all possible permutations or the global maximum number of permutations
    :type total_num_permutations: Optional[int]
    :param num_jobs: number of jobs, defaults to -1
    :type num_jobs: int
    :param method: method to compute the p-value, defaults to "auto".
        "`auto`": if the number of permutations is greater than the maximum number of permutations, the method will be set to "approximate". Otherwise, the method will be set to "exact".
        "`conservative`": p-value is computed as (number of permutations greater or equal than the observed statistic + 1) / (number of permutations + 1).
        "`exact`": p-value is computed as the mean of the binomial cumulative distribution function as stated :cite:`phipson2010permutation`.
        "`approximate`": p-value is computed using the integral of the binomial cumulative distribution function as stated :cite:`phipson2010permutation`.
        "`estimate`": p-value is computed as the mean of the extreme statistic. p-value can be zero.
    :type method: str
    :param random_state: random state, defaults to None
    :type random_state: Optional[int]
    :param verbose: verbose flag, defaults to False
    :type verbose: bool
    :param name: name value, defaults to None. If None, the name will be set to `PermutationTestDistanceBased`.
    :type name: Optional[str]

    :Note:
    Callbacks logs are updated with the following variables:

    - `observed_statistic`: observed statistic obtained from the distance-based detector. Same distance value returned by the `compare` method
    - `permutation_statistic`: list of statistics obtained from the permutations
    - `p_value`: p-value obtained from the permutation test

    :References:

    .. [phipson2010permutation] Phipson, Belinda, and Gordon K. Smyth.
        "Permutation P-values should never be zero: calculating exact P-values when permutations are randomly drawn."
        Statistical applications in genetics and molecular biology 9.1 (2010).

    :Example:

    >>> from frouros.callbacks import PermutationTestDistanceBased
    >>> from frouros.detectors.data_drift import MMD
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> X = np.random.multivariate_normal(mean=[1, 1], cov=[[2, 0], [0, 2]], size=100)
    >>> Y = np.random.multivariate_normal(mean=[0, 0], cov=[[2, 1], [1, 2]], size=100)
    >>> detector = MMD(callbacks=PermutationTestDistanceBased(num_permutations=1000, random_state=31))
    >>> _ = detector.fit(X=X)
    >>> distance, callbacks_log = detector.compare(X=Y)
    >>> distance
    DistanceResult(distance=0.05643613752975596)
    >>> callbacks_log["PermutationTestDistanceBased"]["p_value"]
    0.0009985010823343311
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107  # pylint: disable=too-many-arguments
        self,
        num_permutations: int,
        total_num_permutations: Optional[int] = None,
        num_jobs: int = -1,
        method: str = "auto",
        random_state: Optional[int] = None,
        verbose: bool = False,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(name=name)
        self.num_permutations = num_permutations
        self.total_num_permutations = total_num_permutations
        self.num_jobs = num_jobs
        self.method = method
        self.random_state = random_state
        self.verbose = verbose

    @property
    def num_permutations(self) -> int:
        """Number of permutations property.

        :return: number of permutation to obtain the p-value
        :rtype: int
        """
        return self._num_permutations

    @num_permutations.setter
    def num_permutations(self, value: int) -> None:
        """Number of permutations method setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 1:
            raise ValueError("value must be greater of equal than 1.")
        if value > MAX_NUM_PERM:
            raise ValueError(f"value must be less than or equal to {MAX_NUM_PERM}.")
        self._num_permutations = value

    @property
    def total_num_permutations(self) -> Optional[int]:
        """Number of total permutations' property.

        :return: number of total permutations
        :rtype: Optional[int]
        """
        return self._total_permutations

    @total_num_permutations.setter
    def total_num_permutations(self, value: Optional[int]) -> None:
        """Number of total permutations method setter.

        :param value: value to be set
        :type value: Optional[int]
        :raises ValueError: Value error exception
        """
        if value is not None:
            if value < 1:
                raise ValueError("value must be greater of equal than 1.")
            if value > MAX_NUM_PERM:
                raise ValueError(f"value must be less than or equal to {MAX_NUM_PERM}.")
        self._total_permutations = value

    @property
    def num_jobs(self) -> int:
        """Number of jobs property.

        :return: number of jobs to use
        :rtype: int
        """
        return self._num_jobs

    @num_jobs.setter
    def num_jobs(self, value: int) -> None:
        """Number of jobs method setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value == 0 or value < -1:
            raise ValueError("value must be greater than 0 or -1.")
        self._num_jobs = multiprocessing.cpu_count() if value == -1 else value

    @property
    def method(self) -> str:
        """Method to compute the p-value property.

        :return: method to compute the p-value
        :rtype: str
        """
        return self._method

    @method.setter
    def method(self, value: str) -> None:
        """Method to compute the p-value setter.

        :param value: value to be set
        :type value: str
        :raises ValueError: Value error exception
        """
        if value not in ["auto", "conservative", "exact", "approximate", "estimate"]:
            raise ValueError(
                "value must be one of "
                "'auto', 'conservative', 'exact', 'approximate', 'estimate'.",
            )
        self._method = value

    @property
    def verbose(self) -> bool:
        """Verbose flag property.

        :return: verbose flag
        :rtype: bool
        """
        return self._verbose

    @verbose.setter
    def verbose(self, value: bool) -> None:
        """Verbose flag setter.

        :param value: value to be set
        :type value: bool
        :raises TypeError: Type error exception
        """
        if not isinstance(value, bool):
            raise TypeError("value must of type bool.")
        self._verbose = value

    @staticmethod
    def _calculate_p_value(  # pylint: disable=too-many-arguments
        X_ref: np.ndarray,  # noqa: N803
        X_test: np.ndarray,
        statistic: Callable,
        statistic_args: dict[str, Any],
        observed_statistic: float,
        num_permutations: int,
        total_num_permutations: Optional[int],
        num_jobs: int,
        method: str,
        random_state: Optional[int],
        verbose: bool,
    ) -> Tuple[np.ndarray, float]:
        permuted_statistic, max_num_permutations = permutation(
            X=X_ref,
            Y=X_test,
            statistic=statistic,
            statistical_args=statistic_args,
            num_permutations=num_permutations,
            num_jobs=num_jobs,
            random_state=random_state,
            verbose=verbose,
        )
        permuted_statistic = np.array(permuted_statistic)
        extreme_statistic = permuted_statistic >= observed_statistic  # type: ignore

        if total_num_permutations is None:
            # Set the total number of permutations to the maximum number of
            # permutations, the minimum between all possible permutations or
            # the global maximum number of permutations
            total_num_permutations = np.min([max_num_permutations, MAX_NUM_PERM])

        if method == "auto":
            method = "approximate" if num_permutations > MAX_NUM_PERM else "exact"
        if method == "conservative":
            p_value = PermutationTestDistanceBased._compute_conservative(
                num_permutations=num_permutations,
                observed_statistic=observed_statistic,
                permuted_statistic=permuted_statistic,
            )
        elif method == "exact":
            p_value = PermutationTestDistanceBased._compute_exact(
                extreme_statistic=extreme_statistic,
                total_num_permutations=total_num_permutations,
                permuted_statistic=permuted_statistic,
            )
        elif method == "approximate":
            p_value = PermutationTestDistanceBased._compute_approximate(
                extreme_statistic=extreme_statistic,
                total_num_permutations=total_num_permutations,
                permuted_statistic=permuted_statistic,
            )
        else:  # method == "estimate"
            p_value = PermutationTestDistanceBased._compute_estimate(
                extreme_statistic=extreme_statistic,
            )

        return permuted_statistic, p_value

    @staticmethod
    def _compute_estimate(
        extreme_statistic: np.ndarray,
    ) -> float:
        p_value = extreme_statistic.mean()  # type: ignore
        return p_value

    @staticmethod
    def _compute_approximate(
        extreme_statistic: np.ndarray,
        total_num_permutations: int,
        permuted_statistic: np.ndarray,
    ) -> float:
        num_extreme_statistic = extreme_statistic.sum()
        num_permutations = len(permuted_statistic)
        integral, _ = quad(
            func=lambda p: binom.cdf(
                num_extreme_statistic,
                num_permutations,
                p,
            ),
            a=0,
            b=0.5 / total_num_permutations,
        )
        p_value = (num_extreme_statistic + 1) / (
            num_permutations + 1
        ) - 0.5 / total_num_permutations * integral
        return p_value

    @staticmethod
    def _compute_exact(
        extreme_statistic: np.ndarray,
        total_num_permutations: int,
        permuted_statistic: np.ndarray,
    ) -> float:
        num_extreme_statistic = extreme_statistic.sum()
        num_permutations = len(permuted_statistic)
        p = np.arange(1, total_num_permutations + 1) / total_num_permutations
        y = binom.cdf(
            num_extreme_statistic,
            num_permutations,
            p,
        )
        p_value = np.mean(y)
        return p_value

    @staticmethod
    def _compute_conservative(
        num_permutations: int,
        observed_statistic: float,
        permuted_statistic: np.ndarray,
    ) -> float:
        p_value = (
            (permuted_statistic >= observed_statistic).sum() + 1
        ) / (  # type: ignore
            num_permutations + 1
        )
        return p_value

    def on_compare_end(
        self,
        result: Any,
        X_ref: np.ndarray,  # noqa: N803
        X_test: np.ndarray,
    ) -> None:
        """On compare end method.

        :param result: result obtained from the `compare` method
        :type result: Any
        :param X_ref: reference data
        :type X_ref: numpy.ndarray
        :param X_test: test data
        :type X_test: numpy.ndarray
        """
        observed_statistic = result.distance
        permuted_statistics, p_value = self._calculate_p_value(
            X_ref=X_ref,
            X_test=X_test,
            statistic=self.detector.statistical_method,  # type: ignore
            statistic_args=self.detector.statistical_kwargs,  # type: ignore
            observed_statistic=observed_statistic,
            num_permutations=self.num_permutations,
            total_num_permutations=self.total_num_permutations,
            num_jobs=self.num_jobs,
            method=self.method,
            random_state=self.random_state,
            verbose=self.verbose,
        )
        self.logs.update(
            {
                "observed_statistic": observed_statistic,
                "permuted_statistics": permuted_statistics,
                "p_value": p_value,
            },
        )

    # FIXME: set_detector method as a workaround to  # pylint: disable=fixme
    #  avoid circular import problem. Make it an abstract method and
    #  uncomment commented code when it is solved

    # def set_detector(self, detector) -> None:
    #     """Set detector method.
    #
    #     :raises TypeError: Type error exception
    #     """
    #     if not isinstance(detector, BaseDataDriftBatch):
    #         raise TypeError(
    #             f"callback {self.__class__.name} cannot be used with detector"
    #             f" {detector.__class__name}. Must be used with a detector of "
    #             f"type BaseDataDriftBatch."
    #         )
    #     self.detector = detector

    def reset(self) -> None:
        """Reset method."""



"""Package setup file."""

import codecs

from setuptools import find_packages, setup
import toml


def long_description():
    """Read and return README as long description."""
    with codecs.open("README.md", encoding="utf-8-sig") as f:
        return f.read()


# ground truth package metadata is loaded from pyproject.toml
# for context see:
#   - [PEP 621 -- Storing project metadata in pyproject.toml]
#     (https://www.python.org/dev/peps/pep-0621)
pyproject = toml.load("pyproject.toml")


def setup_package():
    """Set up package."""
    setup(
        author_email=pyproject["project"]["authors"][0]["email"],
        author=pyproject["project"]["authors"][0]["name"],
        description=pyproject["project"]["description"],
        extras_require=pyproject["project"]["optional-dependencies"],
        include_package_data=True,
        install_requires=pyproject["project"]["dependencies"],
        keywords=pyproject["project"]["keywords"],
        classifiers=pyproject["project"]["classifiers"],
        # FIXME: Find a way not to hardcode license
        license="BSD-3-Clause",
        long_description=long_description(),
        long_description_content_type="text/markdown",
        maintainer_email=pyproject["project"]["maintainers"][0]["email"],
        maintainer=pyproject["project"]["maintainers"][0]["name"],
        name=pyproject["project"]["name"],
        packages=find_packages(
            where=".",
            exclude=["tests", "tests.*"],
        ),
        project_urls=pyproject["project"]["urls"],
        python_requires=pyproject["project"]["requires-python"],
        setup_requires=pyproject["build-system"]["requires"],
        url=pyproject["project"]["urls"]["repository"],
        version=pyproject["project"]["version"],
        zip_safe=False,
    )


if __name__ == "__main__":
    setup_package()
